from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.utilities import appendPoint, distBetweenThetas, assignValueIfBetweenRangeForDualDirection, calculateAngle
import zebrazoom.code.util as util
import numpy as np
import math
import cv2

def dualDirectionTailDetection(headPosition, frame, points, angle, maxDepth, steps, nbList, hyperparameters, debug, lenX, lenY):
  
  x = headPosition[0]
  y = headPosition[1]
  
  depth = 0
  angleOpp = (angle + math.pi) % (2 * math.pi)
  
  lastFirstTheta = angle
  
  pixSurMax = hyperparameters["headEmbededParamTailDescentPixThreshStop"]
  
  pixTotList = []
  
  distSubsquentPoints = 0.000001
  pixSur = 0
  
  if debug:
    frameDisplay = frame.copy()
    cv2.circle(frameDisplay, (x, y), 3, (0,0,0),   -1)
    util.showFrame(frameDisplay, title="HeadEmbeddedTailTracking")
  
  if ("minimumPixelIntensitySumForAnimalDetection" in hyperparameters) and (np.sum(np.sum(255 - frame)) < hyperparameters["minimumPixelIntensitySumForAnimalDetection"]):
    return ([], angle, 255)
  
  while (distSubsquentPoints > 0 and depth < maxDepth and ((pixSur < pixSurMax) or (depth < hyperparameters["authorizedRelativeLengthTailEnd"]*maxDepth))):
  
    if depth == 0 and angle == -99999: # First frame, first segment
      thetaDiffAccept = 1.8
    else:
      if (depth < hyperparameters["authorizedRelativeLengthTailEnd"]*maxDepth):
        thetaDiffAccept = hyperparameters["thetaDiffAccept"]
      else:
        if not("authorizedRelativeLengthTailEnd2" in hyperparameters) or (depth < hyperparameters["authorizedRelativeLengthTailEnd2"]*maxDepth):
          thetaDiffAccept = hyperparameters["thetaDiffAcceptAfterAuthorizedRelativeLengthTailEnd"]
          nbList = hyperparameters["nbListAfterAuthorizedRelativeLengthTailEnd"]
        else:
          thetaDiffAccept = hyperparameters["thetaDiffAcceptAfterAuthorizedRelativeLengthTailEnd2"]
          nbList = hyperparameters["nbListAfterAuthorizedRelativeLengthTailEnd2"]
    
    # if debug:
      # print("thetaDiffAccept:", thetaDiffAccept)
    
    # Initial direction
    if len(points[0]) == 0:
      xOld = x
      yOld = y
    else:
      nbCols = len(points[0])
      xOld = points[0][nbCols - 1]
      yOld = points[1][nbCols - 1]
    pixTotMax = 1000000
    maxTheta  = angle
    l = [i*(math.pi/nbList) for i in range(0,2*nbList) if distBetweenThetas(i*(math.pi/nbList), angle) < thetaDiffAccept]
    if debug:
      print("l normal:", l)
    
    xTot = -1
    yTot = -1
    pixSur = -1
    for step in steps:
      if (step < maxDepth - depth) or (step == steps[0]):
        for theta in l:
          xNew = assignValueIfBetweenRangeForDualDirection(int(xOld + step * (math.cos(theta))), 0, lenX)
          yNew = assignValueIfBetweenRangeForDualDirection(int(yOld + step * (math.sin(theta))), 0, lenY)
          pixTot = frame[yNew][xNew]
          if xNew != -1 and yNew != -1 and (pixTot < pixTotMax):
            pixTotMax = pixTot
            maxTheta = theta
            xTot = xNew
            yTot = yNew
            if debug:
              print("regular dir: Choosing (x, y):", xTot, yTot, "; step:", step, "; theta:", theta, "; pixTot:", pixTot)
    if xTot != -1 and yTot != -1:
      pixTotList.append(pixTotMax)
      pixSur = frame[yTot, xTot]

    # Opposite direction
    if len(points[0]) == 0:
      xOld = x
      yOld = y
    else:
      xOld = points[0][0]
      yOld = points[1][0]
    pixTotMax = 1000000 # remove this line?
    maxTheta  = angleOpp
    l = [i*(math.pi/nbList) for i in range(0,2*nbList) if distBetweenThetas(i*(math.pi/nbList), angleOpp) < thetaDiffAccept]
    xTotOpp = -1
    yTotOpp = -1
    pixSurOpp = -1
    if debug:
      print("l opposite:", l)
    for step in steps:
      if (step < maxDepth - depth) or (step == steps[0]):
        for theta in l:
          xNew = assignValueIfBetweenRangeForDualDirection(int(xOld + step * (math.cos(theta))), 0, lenX)
          yNew = assignValueIfBetweenRangeForDualDirection(int(yOld + step * (math.sin(theta))), 0, lenY)
          pixTot = frame[yNew][xNew]
          if xNew != -1 and yNew != -1 and (pixTot < pixTotMax):
            pixTotMax = pixTot
            maxTheta = theta
            xTotOpp = xNew
            yTotOpp = yNew
            if debug:
              print("opposite dir: Choosing (x, y):", xTotOpp, yTotOpp, "; step:", step, "; theta:", theta, "; pixTot:", pixTot)
    if xTotOpp != -1 and yTotOpp != -1:
      pixTotList.append(pixTotMax)
      pixSurOpp = frame[yTotOpp, xTotOpp]
    
    # Choosing between initial and opposite direction
    if pixSur != -1 and pixSurOpp != -1:
      if pixSurOpp < pixSur:
        xTot   = xTotOpp
        yTot   = yTotOpp
        pixSur = pixSurOpp
        oppChosen = True
      else:
        oppChosen = False
    elif pixSur != -1:
      oppChosen = False
    elif pixSurOpp != -1:
      xTot   = xTotOpp
      yTot   = yTotOpp
      pixSur = pixSurOpp
      oppChosen = True
    else:
      xTot = -1
      yTot = -1
      oppChosen = False
      
    # if debug:
      # print("oppChosen:", oppChosen)
      # print("xTot, yTot:", xTot, yTot)
    
    if xTot != -1 and yTot != -1:
      # Calculates distance between new and old point
      if len(points[0]) == 0:
        xOld = x
        yOld = y
      else:
        if oppChosen:
          xOld = points[0][0]
          yOld = points[1][0]
        else:
          nbCols = len(points[0])
          xOld = points[0][nbCols - 1]
          yOld = points[1][nbCols - 1]
      
      distSubsquentPoints = math.sqrt((xTot - xOld)**2 + (yTot - yOld)**2)
      # if debug:
        # print("distSubsquentPoints:", distSubsquentPoints)
      
      if depth + distSubsquentPoints < maxDepth and ((pixSur < pixSurMax) or (depth < hyperparameters["authorizedRelativeLengthTailEnd"]*maxDepth)):
        if oppChosen:
          points = np.insert(points, 0, [xTot, yTot], 1)
        else:
          points = appendPoint(xTot, yTot, points)
      else:
        # if debug:
          # print("diff")
        vectX = xTot - xOld
        vectY = yTot - yOld
        xTot  = int(xOld + (maxDepth / (depth + distSubsquentPoints)) * vectX)
        yTot  = int(yOld + (maxDepth / (depth + distSubsquentPoints)) * vectY)
        if oppChosen:
          points = np.insert(points, 0, [xTot, yTot], 1)
        else:
          points = appendPoint(xTot, yTot, points)
      
      if debug:
        print("points:", points)
        cv2.circle(frameDisplay, (xTot, yTot), 1, (0,0,0),   -1)
        util.showFrame(frameDisplay, title="HeadEmbeddedTailTracking")
      
      newTheta = calculateAngle(xOld,yOld,xTot,yTot)
      
      if depth == 0:
        if oppChosen:
          angleOpp = newTheta
          angle    = (newTheta + math.pi) % (2 * math.pi)
        else:
          angle    = newTheta 
          angleOpp = (newTheta + math.pi) % (2 * math.pi)
      else:
        if oppChosen:
          angleOpp = newTheta
        else:
          angle = newTheta
      
      if depth == 0:
        lastFirstTheta = angle
      depth = depth + distSubsquentPoints
      
      if xTot == -1 or yTot == -1:
        distSubsquentPoints = 0
      
      if debug:
        print("distSubsquentPoints:", distSubsquentPoints, "; depth < maxDepth:", depth < maxDepth, "; pixSur < pixSurMax:", pixSur < pixSurMax, '; depth < hyperparameters["authorizedRelativeLengthTailEnd"]*maxDepth):', depth < hyperparameters["authorizedRelativeLengthTailEnd"]*maxDepth)
    
    else:
      distSubsquentPoints = 0
  
  lenPoints = len(points[0]) - 1
  if points[0, lenPoints-1] == points[0, lenPoints] and points[1, lenPoints-1] == points[1, lenPoints]:
    points = points[:, :len(points[0])-1]
  
  # if debug:
    # print(np.median(pixTotList), np.mean(pixTotList), pixTotList)
  
  medianPixTotList = np.median(pixTotList)
  
  if ("dualDirectionRemoveShortestDirectionFromHead" in hyperparameters) and hyperparameters["dualDirectionRemoveShortestDirectionFromHead"]:
    accumulatedLength1 = 0
    accumulatedLength2 = 0
    headPositionIndex = -1
    passedHeadPosition = False
    for i in range(len(points[0])):
      if i > 0:
        length = math.sqrt((points[0, i-1] - points[0, i])**2 + (points[1, i-1] - points[1, i])**2)
        if not(passedHeadPosition):
          accumulatedLength1 += length
        else:
          accumulatedLength2 += length
      if (headPosition[0] == points[0, i]) and (headPosition[1] == points[1, i]):
        passedHeadPosition = True
        headPositionIndex  = i
    if accumulatedLength1 > accumulatedLength2:
      points2 = points[:, :headPositionIndex]
      points  = points2[:, ::-1]
    else:
      points  = points[:, headPositionIndex+1:]
  
  return (points, lastFirstTheta, medianPixTotList)
  


from sklearn.svm import SVC
from sklearn.neighbors import KNeighborsClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.cluster import KMeans
from sklearn.mixture import GaussianMixture
from zebrazoom.dataAnalysis.dataanalysis.visualizeClusters import visualizeClusters
import matplotlib.pyplot as plt
import numpy as np
import pickle
import shutil
import json
import os
import sys


def prepareForActiveLearning(proportions, sortedRepresentativeBouts, outputFolderResult, nbCluster, pca_result, dfParam, sortedRepresentativeBoutsIndex, tailAngles):
  
  outputFolderResult2 = os.path.join(outputFolderResult, 'activeLearning')
  if os.path.exists(outputFolderResult2):
    shutil.rmtree(outputFolderResult2)
  os.mkdir(outputFolderResult2)
  
  optimalLength = 12
  nbIterations  = 5
  
  # Plotting bouts one by one for each cluster and writing bout indices in txt file
  outF = open(os.path.join(outputFolderResult2, 'boutIndices.txt'), "w")
  for classed in range(0, len(proportions[0])):
    indices = sortedRepresentativeBouts[classed].index
    beginningMiddleEnd = False
    if beginningMiddleEnd:
      for j in range(0, 3):
        ind2size = optimalLength if len(indices) >= 3*optimalLength else max(1, int(len(indices) / 3))
        if j == 0:
          indices2 = indices[0:ind2size]
        elif j == 1:
          indices2 = indices[int(len(indices)/2) - int(ind2size/2) : int(len(indices)/2) + int(ind2size/2)]
        else:
          indices2 = indices[len(indices) - ind2size : len(indices)]
        # Plot
        fig, tabAx3 = plt.subplots(4, 3, figsize=(22.9, 8.8))
        for k in range(0, len(indices2)):
          tailAnglestab = dfParam.loc[indices2[k], tailAngles].tolist()
          tabAx3[int(k / 3), k % 3].plot(tailAnglestab, 'b', label = str(indices2[k]))
          tabAx3[int(k / 3), k % 3].legend()
          tabAx3[int(k / 3), k % 3].title.set_text(str(indices2[k]))
        region = 'mostRepresentative' if j == 0 else 'inBetween' if j == 1 else 'leastRepresentative'
        plt.savefig(os.path.join(outputFolderResult2, 'cluster' + str(classed + 1) + '_' + region + '.png'))
        # Saving indices in txt
        line = 'cluster' + str(classed) + '_' + region + ' = [' + ', '.join([str(ind) for ind in indices2]) + "]\n"
        outF.write(line)
    else:
      for j in range(0, nbIterations):
        ind2size = optimalLength if len(indices) >= 3*optimalLength else max(1, int(len(indices) / 3))
        if (j+1)*ind2size < len(indices):
          indices2 = indices[j*ind2size:(j+1)*ind2size]
          # Plot
          fig, tabAx3 = plt.subplots(4, 3, figsize=(22.9, 8.8))
          for k in range(0, len(indices2)):
            tailAnglestab = dfParam.loc[indices2[k], tailAngles].tolist()
            tabAx3[int(k / 3), k % 3].plot(tailAnglestab, 'b', label = str(indices2[k]))
            tabAx3[int(k / 3), k % 3].legend()
            tabAx3[int(k / 3), k % 3].title.set_text(str(indices2[k]))
          region = 'mostRepresentative'
          plt.savefig(os.path.join(outputFolderResult2, 'cluster' + str(classed + 1) + '_' + region + '_' + str(j*ind2size) + '_' + str((j+1)*ind2size) + '.png'))
          # Saving indices in txt
          line = 'cluster' + str(classed) + '_' + region + '_' + str(j*ind2size) + '_' + str((j+1)*ind2size) + ' = [' + ', '.join([str(ind) for ind in indices2]) + "]\n"
          outF.write(line)
      for j in range(0, nbIterations):
        ind2size = optimalLength if len(indices) >= 3*optimalLength else max(1, int(len(indices) / 3))
        if len(indices)-(j+1)*ind2size > 0:
          indices2 = indices[len(indices)-(j+1)*ind2size:len(indices)-j*ind2size]
          # Plot
          fig, tabAx3 = plt.subplots(4, 3, figsize=(22.9, 8.8))
          for k in range(0, len(indices2)):
            tailAnglestab = dfParam.loc[indices2[k], tailAngles].tolist()
            tabAx3[int(k / 3), k % 3].plot(tailAnglestab, 'b', label = str(indices2[k]))
            tabAx3[int(k / 3), k % 3].legend()
            tabAx3[int(k / 3), k % 3].title.set_text(str(indices2[k]))
          region = 'leastRepresentative'
          plt.savefig(os.path.join(outputFolderResult2, 'cluster' + str(classed + 1) + '_' + region + '_' + str(len(indices)-(j+1)*ind2size) + '_' + str(len(indices)-j*ind2size) + '.png'))
          # Saving indices in txt
          line = 'cluster' + str(classed) + '_' + region + '_' + str(len(indices)-(j+1)*ind2size) + '_' + str(len(indices)-j*ind2size) + ' = [' + ', '.join([str(ind) for ind in indices2]) + "]\n"
          outF.write(line)
  outF.close()
  
  pickle.dump({'pca_result': pca_result.tolist(), 'dfParam': dfParam}, open(os.path.join(outputFolderResult2, 'boutParameters.pkl'), 'wb'))


def activeLearning(modelUsed, nbConditions, nbCluster, outputFolderResult, N_QUERIES, manualClassicationPath):
  from PyQt5.QtWidgets import QInputDialog

  from modAL.models import ActiveLearner
  
  nbFramesTakenIntoAccount = 24
  scaleGraphs   = 1
  showFigures   = 0
  
  with open(os.path.join(outputFolderResult, 'boutParameters.pkl'), 'rb') as pickle_file:
    data = pickle.load(pickle_file)
  with open(manualClassicationPath, 'rb') as classification_file:
    manualClassifications = json.load(classification_file)
  
  pca_result = np.array(data['pca_result'])
  dfParam    = data['dfParam']

  y_train    = np.array([])
  ind_train  = []
  for ind in manualClassifications:
    classIndIndices = manualClassifications[ind]
    ind_train       = ind_train + classIndIndices
    y_train         = np.concatenate((y_train, np.array([int(ind) for i in range(0, len(classIndIndices))])))
  X_train = pca_result[ind_train, :]
  
  X_pool   = np.delete(pca_result, ind_train, axis=0)
  ind_pool = np.delete(np.array([ind for ind in range(0, len(pca_result))]), ind_train, axis=0)
  tailAngles = ['tailAngles' + str(i) for i in range(1, 25)]
  tailAngles_pool = np.array(dfParam.loc[ind_pool, tailAngles])
  
  # Learning (and predicting) first model
  
  if modelUsed == 'KNeighborsClassifier':
    model = KNeighborsClassifier(n_neighbors = nbCluster)
  elif modelUsed == 'SVC':
    model = SVC()
  elif modelUsed == 'GaussianNB':
    model = GaussianNB()
  elif modelUsed == 'KMeans':
    model = KMeans(n_clusters = nbCluster)
  elif modelUsed == 'GaussianMixture':
    model = GaussianMixture(n_components = nbCluster)
  else:
    model = KMeans(n_clusters = nbCluster)
  
  learner = ActiveLearner(estimator=model, X_training=X_train, y_training=y_train)
  
  predictions = learner.predict(pca_result)
  
  print("Prediction after first training:")
  print(predictions)
  
  # Actual active learning start
  
  for index in range(N_QUERIES):
    
    query_index, query_instance = learner.query(X_pool)
    
    plt.plot(tailAngles_pool[query_index][0])
    plt.show()
    
    userInput, _ok = QInputDialog.getInt(None, "Cluster Number", "Enter the cluster number of this bout:")
    
    X, y = X_pool[query_index].reshape(1, -1), np.array([userInput])
    
    learner.teach(X=X, y=y)
    
    # Remove the queried instance from the unlabeled pool. # Add tail angle remove
    X_pool = np.delete(X_pool, query_index, axis=0)
    tailAngles_pool = np.delete(tailAngles_pool, query_index, axis=0)
  
  # See new results
  
  predictions = learner.predict(pca_result)
  
  print("Prediction after second training:")
  print(predictions)
  
  outputFolderResult3 = os.path.join(outputFolderResult, 'activeLearnOutput' + str(1))
  if os.path.exists(outputFolderResult3):
    shutil.rmtree(outputFolderResult3)
  os.mkdir(outputFolderResult3)
  
  predictedProb = []
  
  visualizeClusters(dfParam, predictions, predictedProb, modelUsed, nbConditions, nbCluster, nbFramesTakenIntoAccount, scaleGraphs, showFigures, outputFolderResult3, 0, 1)
  
  


from zebrazoom.code.extractParameters import extractParameters

import cv2
import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
import numpy as np
import math
from zebrazoom.code.tracking import register_tracking_method
from ._eyeTracking import EyeTrackingMixin
from ._fasterMultiprocessingBase import BaseFasterMultiprocessing
from ._getImages import GetImagesMixin


class FasterMultiprocessing(BaseFasterMultiprocessing, EyeTrackingMixin, GetImagesMixin):
  def __init__(self, videoPath, wellPositions, hyperparameters):
    super().__init__(videoPath, wellPositions, hyperparameters)
    if self._hyperparameters["eyeTracking"]:
      self._trackingEyesAllAnimalsList = [np.zeros((self._hyperparameters["nbAnimalsPerWell"], self._lastFrame-self._firstFrame+1, 8))
                                          for _ in range(self._hyperparameters["nbWells"])]
    else:
      self._trackingEyesAllAnimals = 0

    if not(self._hyperparameters["nbAnimalsPerWell"] > 1) and not(self._hyperparameters["headEmbeded"]) and (self._hyperparameters["findHeadPositionByUserInput"] == 0) and (self._hyperparameters["takeTheHeadClosestToTheCenter"] == 0):
      self._trackingProbabilityOfGoodDetectionList = [np.zeros((self._hyperparameters["nbAnimalsPerWell"], self._lastFrame-self._firstFrame+1))
                                                      for _ in range(self._hyperparameters["nbWells"])]
    else:
      self._trackingProbabilityOfGoodDetectionList = 0

  def _adjustParameters(self, i, frame, widgets):
    return None

  def _formatOutput(self):
    if self._hyperparameters["postProcessMultipleTrajectories"]:
      for wellNumber in range(self._firstWell, self._lastWell + 1):
        self._postProcessMultipleTrajectories(self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingProbabilityOfGoodDetectionList[wellNumber])

    return {wellNumber: extractParameters([self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingHeadingAllAnimalsList[wellNumber], [], 0, 0] + ([self._auDessusPerAnimalIdList[wellNumber]] if self._auDessusPerAnimalIdList is not None else []), wellNumber, self._hyperparameters, self._videoPath, self._wellPositions, self._background)
            for wellNumber in range(self._firstWell, self._lastWell + 1)}

  def run(self):
    self._background = self.getBackground()

    cap = zzVideoReading.VideoCapture(self._videoPath)
    if (cap.isOpened()== False):
      print("Error opening video stream or file")

    if self._hyperparameters["backgroundSubtractorKNN"]:
      fgbg = cv2.createBackgroundSubtractorKNN()
      for i in range(0, min(self._lastFrame - 1, 500), int(min(self._lastFrame - 1, 500) / 10)):
        cap.set(1, min(self._lastFrame - 1, 500) - i)
        ret, frame = cap.read()
        fgmask = fgbg.apply(frame)
      cap.release()
      cap = zzVideoReading.VideoCapture(self._videoPath)

    i = self._firstFrame

    if self._firstFrame:
      cap.set(1, self._firstFrame)

    previousFrames = None
    widgets = None
    while (i < self._lastFrame + 1):

      if (self._hyperparameters["freqAlgoPosFollow"] != 0) and (i % self._hyperparameters["freqAlgoPosFollow"] == 0):
        print("Tracking: frame:",i)
        if self._hyperparameters["popUpAlgoFollow"]:
          from zebrazoom.code.popUpAlgoFollow import prepend

          prepend("Tracking: frame:" + str(i))

      if self._hyperparameters["debugTracking"]:
        print("frame:",i)

      ret, frame = cap.read()

      if ret:

        if self._hyperparameters["backgroundSubtractorKNN"]:
          frame = fgbg.apply(frame)
          frame = 255 - frame

        for wellNumber in range(self._firstWell, self._lastWell + 1):
          if self._hyperparameters["nbAnimalsPerWell"] == 1 and not(self._hyperparameters["forceBlobMethodForHeadTracking"]):
            minPixelDiffForBackExtract = self._hyperparameters["minPixelDiffForBackExtract"]
            xtop = self._wellPositions[wellNumber]['topLeftX']
            ytop = self._wellPositions[wellNumber]['topLeftY']
            lenX = self._wellPositions[wellNumber]['lengthX']
            lenY = self._wellPositions[wellNumber]['lengthY']
            if self._hyperparameters["backgroundSubtractorKNN"]:
              grey = frame
            else:
              grey = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

            curFrame = grey[ytop:ytop+lenY, xtop:xtop+lenX].copy()
            if not(self._hyperparameters["backgroundSubtractorKNN"]):
              back = self._background[ytop:ytop+lenY, xtop:xtop+lenX]
              putToWhite = ( curFrame.astype('int32') >= (back.astype('int32') - minPixelDiffForBackExtract) )
              curFrame[putToWhite] = 255
            else:
              self._hyperparameters["paramGaussianBlur"] = int(math.sqrt(cv2.countNonZero(255 - curFrame) / self._hyperparameters["nbAnimalsPerWell"]) / 2) * 2 + 1
            if self._hyperparameters["paramGaussianBlur"]:
              blur = cv2.GaussianBlur(curFrame, (self._hyperparameters["paramGaussianBlur"], self._hyperparameters["paramGaussianBlur"]),0)
            else:
              blur = curFrame
            thresh1 = 0
            thresh2 = 0
            gray    = 0
          else:
            [frame2, gray, thresh1, blur, thresh2, frame2, initialCurFrame, back, xHead, yHead] = self._getImages(0, i, wellNumber, frame)

          headPositionFirstFrame = 0

          # Head tracking and heading calculation
          lastFirstTheta = self._headTrackingHeadingCalculation(i, blur, thresh1, thresh2, gray, self._hyperparameters["erodeSize"], int(cap.get(3)), int(cap.get(4)), self._trackingHeadingAllAnimalsList[wellNumber], self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingProbabilityOfGoodDetectionList[wellNumber], headPositionFirstFrame, self._wellPositions[wellNumber]["lengthX"])

          # Tail tracking for frame i
          if self._hyperparameters["trackTail"] == 1 :
            threshForBlackFrames = 0
            thetaDiffAccept = 1.2
            lastFirstTheta = 0
            maxDepth = 0
            tailTipFirstFrame = []
            for animalId in range(0, self._hyperparameters["nbAnimalsPerWell"]):
              self._tailTracking(animalId, i, frame, thresh1, threshForBlackFrames, thetaDiffAccept, self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingHeadingAllAnimalsList[wellNumber], lastFirstTheta, maxDepth, tailTipFirstFrame, initialCurFrame.copy(), back)

          # Eye tracking for frame i
          if self._hyperparameters["eyeTracking"]:
            self._eyeTracking(animalId, i, frame, thresh1, self._trackingHeadingAllAnimalsList[wellNumber], self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingHeadingAllAnimalsList[wellNumber])

          self._debugTracking(i, self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingHeadingAllAnimalsList[wellNumber], curFrame)

          if self._hyperparameters["freqAlgoPosFollow"]:
            if i % self._hyperparameters["freqAlgoPosFollow"] == 0:
              print("Tracking at frame", i)

        if self._hyperparameters["detectMovementWithRawVideoInsideTracking"]:
          previousFrames = self._detectMovementWithRawVideoInsideTracking(i, grey, previousFrames)

      paramsAdjusted = self._adjustParameters(i, frame, widgets)
      if paramsAdjusted is not None:
        i, widgets = paramsAdjusted
        cap.set(1, i)
      else:
        i = i + 1

    cap.release()
    return self._formatOutput()


register_tracking_method('fasterMultiprocessing', FasterMultiprocessing)



import math
import os

import cv2
import numpy as np
from numpy import linspace
from scipy.interpolate import interp1d

import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading

from ._tailTrackingBase import TailTrackingBase


class HeadEmbeddedTailTrackingTeresaNicolsonMixin(TailTrackingBase):
  @staticmethod
  def __smoothTail(points, nbTailPoints):
    y = points[0]
    x = linspace(0, 1, len(y))

    if len(x) > 3:

      points2 = []
      for i in range(0, len(points[0])-1):
        if not((points[0][i] == points[0][i+1]) and (points[1][i] == points[1][i+1])):
          points2.append([points[0][i], points[1][i]])

      i = len(points[0]) - 1
      if not((points[0][i-1] == points[0][i]) and (points[1][i-1] == points[1][i])):
        points2.append([points[0][i], points[1][i]])

      points = np.array(points2).T

      # Define some points:
      points = np.array([points[0], points[1]]).T  # a (nbre_points x nbre_dim) array

      # Linear length along the line:
      distance = np.cumsum( np.sqrt(np.sum( np.diff(points, axis=0)**2, axis=1 )) )
      distance = np.insert(distance, 0, 0)/distance[-1]

      # Interpolation for different methods:
      interpolation_method = 'quadratic'    # 'slinear', 'quadratic', 'cubic'
      alpha = np.linspace(0, 1, nbTailPoints)

      interpolated_points = {}

      interpolator =  interp1d(distance, points, kind=interpolation_method, axis=0)
      interpolated_points = interpolator(alpha)

      interpolated_points = interpolated_points.T

      newX = interpolated_points[0]
      newY = interpolated_points[1]

    else:

      newX = points[0]
      newY = points[1]

    return [newX, newY]

  def __findNextPoints(self, depth, x, y, frame, points, angle, maxDepth, steps, nbList, initialImage, debug):
    lenX = len(frame[0]) - 1
    lenY = len(frame) - 1

    thetaDiffAccept = 1 #0.4

    if depth < 0.15*maxDepth:
      thetaDiffAccept = 0.4

    if depth > 0.85*maxDepth:
      thetaDiffAccept = 0.6

    pixTotMax = 1000000
    maxTheta  = angle

    l = [i*(math.pi/nbList) for i in range(0,2*nbList) if self._distBetweenThetas(i*(math.pi/nbList), angle) < thetaDiffAccept]

    if debug:
      print("debug")

    xTot = self._assignValueIfBetweenRange(x + steps[0], 0, lenX)
    yTot = self._assignValueIfBetweenRange(y, 0, lenY)
    if yTot == 0:
      yTot = 400

    for step in steps:

      for theta in l:

        xNew = self._assignValueIfBetweenRange(int(x + step * (math.cos(theta))), 0, lenX)
        yNew = self._assignValueIfBetweenRange(int(y + step * (math.sin(theta))), 0, lenY)
        pixTot = frame[yNew][xNew]

        if debug:
          print([theta,pixTot])

        # Keeps that theta angle as maximum if appropriate
        if (pixTot < pixTotMax):
          pixTotMax = pixTot
          maxTheta = theta
          if depth < 0.4*maxDepth:
            if xNew > x:
              xTot = xNew
              yTot = yNew
          else:
            xTot = xNew
            yTot = yNew

    w = 8 # THIS IS IMPORTANT
    ym = yTot - w
    yM = yTot + w
    xm = xTot - w
    xM = xTot + w
    if ym < 0:
      ym = 0
    if xm < 0:
      xm = 0
    if yM > len(initialImage):
      yM = len(initialImage)
    if xM > len(initialImage[0]):
      xM = len(initialImage[0])

    pixSur = np.min(initialImage[ym:yM, xm:xM])
    if debug:
      print("depth:", depth, " ; maxDepth:", maxDepth, " ; pixSur:", pixSur)

    # if depth > 0.95*maxDepth:
      # pixTot = frame[y][x]
      # if (pixTot < pixTotMax):
        # pixTotMax = pixTot
        # maxTheta = theta
        # xTot = x
        # yTot = y
        # depth = maxDepth + 10

    if debug:
      print(["max:",maxTheta,pixTotMax])

    # Calculates distance between new and old point
    distSubsquentPoints = math.sqrt((xTot - x)**2 + (yTot - y)**2)

    pixSurMax = 150
    if ((pixSur < pixSurMax) or (depth < 2*0.85*maxDepth)):
      points = self._appendPoint(xTot, yTot, points)
      if debug:
        cv2.circle(frame, (xTot, yTot), 3, (255,0,0),   -1)
        self._debugFrame(frame, title='HeadEmbeddedTailTracking')

    newTheta = self._calculateAngle(x,y,xTot,yTot)

    if (distSubsquentPoints > 0) and (depth < 2*maxDepth) and (xTot < 1280 - 10) and (yTot > 10) and (yTot < 1024 - 10) and ((pixSur < pixSurMax) or (depth < 2*0.85*maxDepth)):
      (points,nop) = self.__findNextPoints(depth+distSubsquentPoints,xTot,yTot,frame,points,newTheta,maxDepth,steps,nbList,initialImage,debug)

    return (points,newTheta)

  def _headEmbededTailTrackingTeresaNicolson(self, headPosition, frame, maxDepth, tailTip, threshForBlackFrames):
    steps   = self._hyperparameters["step"]
    nbList  = 10

    x = headPosition[0]
    y = headPosition[1]

    initialImage = frame.copy()

    gaussian_blur = self._hyperparameters["headEmbededParamGaussianBlur"]
    frame = cv2.GaussianBlur(frame, (gaussian_blur, gaussian_blur), 0)
    # angle = self._hyperparameters["headEmbededParamInitialAngle"]
    angle = self._calculateAngle(x, y, tailTip[0], tailTip[1])

    points = np.zeros((2, 0))

    if np.mean(np.mean(frame)) > threshForBlackFrames:
      (points, lastFirstTheta2) = self.__findNextPoints(0,x,y,frame,points,angle,maxDepth,steps,nbList,initialImage,self._hyperparameters["debugHeadEmbededFindNextPoints"])
      points = np.insert(points, 0, headPosition, axis=1)
      if len(points[0]) > 3:
        points = self.__smoothTail(points, self._nbTailPoints)
      points[0][0] = headPosition[0]
      points[1][0] = headPosition[1]
    else:
      points = np.zeros((2, self._nbTailPoints))

    output = np.zeros((1, len(points[0]), 2))

    for idx, x in enumerate(points[0]):
      output[0][idx][0] = x
      output[0][idx][1] = points[1][idx]

    return output

  def _headEmbededTailTrackFindMaxDepthTeresaNicolson(self, frame):
    x = self._headPositionFirstFrame[0]
    y = self._headPositionFirstFrame[1]

    steps   = self._hyperparameters["step"]
    nbList  = 10

    initialImage = frame.copy()

    gaussian_blur = self._hyperparameters["headEmbededParamGaussianBlur"]
    frame = cv2.GaussianBlur(frame, (gaussian_blur, gaussian_blur), 0)

    angle = self._calculateAngle(x, y, self._tailTipFirstFrame[0], self._tailTipFirstFrame[1])

    points = np.zeros((2, 0))

    (points, lastFirstTheta2) = self.__findNextPoints(0,x,y,frame,points,angle,self._hyperparameters["headEmbededTailTrackFindMaxDepthInitialMaxDepth"],steps,nbList,initialImage, self._hyperparameters["debugHeadEmbededFindNextPoints"])

    distToTip        = np.full((200),10000)
    curTailLengthTab = np.full((200),10000)
    curTailLength  = 0
    k = 0

    distFromHeadToTip = abs(math.sqrt((x-self._tailTipFirstFrame[0])**2 + (y-self._tailTipFirstFrame[1])**2))
    while (curTailLength < 1.5*distFromHeadToTip) and (k < len(points[0])-1):
      curTailLength = curTailLength + abs(math.sqrt((points[0,k]-points[0,k+1])**2 + (points[1,k]-points[1,k+1])**2))
      distToTip[k]  = abs(math.sqrt((points[0,k]-self._tailTipFirstFrame[0])**2 + (points[1,k]-self._tailTipFirstFrame[1])**2))
      curTailLengthTab[k] = curTailLength
      k = k + 1

    minDistToTip    = 1000000
    indMinDistToTip = 0
    for idx, dist in enumerate(distToTip):
      if dist < minDistToTip:
        minDistToTip = dist
        indMinDistToTip = idx

    return (curTailLengthTab[indMinDistToTip] )

  def __getMeanOfImageOverVideo(self):
    cap = zzVideoReading.VideoCapture(self._videoPath)
    meanss = []
    ret = True
    i = 0
    while (i < 100):
      ret, frame = cap.read()
      if ret:
        if self._hyperparameters["invertBlackWhiteOnImages"]:
          frame = 255 - frame
        val = np.mean(np.mean(frame))
        meanss.append(val)
      i = i +1
    return np.mean(meanss)

  def _getThresForBlackFrame(self):
    threshForBlackFrames = 0
    if self._hyperparameters["headEmbededTeresaNicolson"] == 1:
      imagesMeans = self.__getMeanOfImageOverVideo()
      threshForBlackFrames = imagesMeans * 0.8 #0.75
    return threshForBlackFrames

  def _savingBlackFrames(self, output):
    if self._hyperparameters["headEmbededTeresaNicolson"] == 1:
      if self._hyperparameters["noBoutsDetection"] == 1:
        outputPath = os.path.join(self._hyperparameters["outputFolder"], self._hyperparameters['videoNameWithTimestamp'])
        if not os.path.exists(outputPath):
          os.makedirs(outputPath)
        with open(os.path.join(outputPath, f'blackFrames_{self._videoName}.csv'), "a") as f:
          for k in range(1,len(output[0])):
            if np.sum(output[0, k]) == 0:
              output[0, k] = output[0, k-1]
              f.write(str(k)+'\n')



import numpy as np
import cv2

def getNewFrameROI(self, k, frame, wellNumber, animalId):

  # Retrieving ROI coordinates and selecting ROI
  roiXStart = self._wellPositions[wellNumber]['topLeftX'] + int(self._trackingDataPerWell[wellNumber][animalId][k][0][0] - self._hyperparameters["backgroundSubtractionOnROIhalfDiameter"])
  roiYStart = self._wellPositions[wellNumber]['topLeftY'] + int(self._trackingDataPerWell[wellNumber][animalId][k][0][1] - self._hyperparameters["backgroundSubtractionOnROIhalfDiameter"])
  roiXEnd   = roiXStart + 2 * self._hyperparameters["backgroundSubtractionOnROIhalfDiameter"]
  roiYEnd   = roiYStart + 2 * self._hyperparameters["backgroundSubtractionOnROIhalfDiameter"]
  if roiXStart < 0:
    roiXStart = 0
    roiXEnd   = 2 * self._hyperparameters["backgroundSubtractionOnROIhalfDiameter"]
  if roiYStart < 0:
    roiYStart = 0
    roiYEnd   = 2 * self._hyperparameters["backgroundSubtractionOnROIhalfDiameter"]
  if roiXEnd >= len(frame[0]):
    roiXEnd   = len(frame[0]) - 1
    roiXStart = len(frame[0]) - 2 * self._hyperparameters["backgroundSubtractionOnROIhalfDiameter"]
  if roiYEnd >= len(frame):
    roiYEnd   = len(frame) - 1
    roiYStart = len(frame) - 2 * self._hyperparameters["backgroundSubtractionOnROIhalfDiameter"]
  frameROI = frame[roiYStart:roiYEnd, roiXStart:roiXEnd].copy()
  
  # Subtracting background of image
  backgroundROI = self._background[roiYStart:roiYEnd, roiXStart:roiXEnd]
  frameROI = 255 - np.where(backgroundROI >= frameROI, backgroundROI - frameROI, 0).astype(np.uint8)

  # Applying gaussian filter
  paramGaussianBlur = self._hyperparameters["paramGaussianBlur"]
  frameROI = cv2.GaussianBlur(frameROI, (paramGaussianBlur, paramGaussianBlur), 0)
  
  return frameROI


import json
import numpy as np
import pandas as pd
import os
import time
import datetime
import math
from difflib import SequenceMatcher

from zebrazoom.dataAPI._createSuperStructFromH5 import createSuperStructFromH5
from zebrazoom.dataAPI._openResultsFile import openResultsFile


def calculateSleepVsMovingPeriods(pathToZZoutput, vidName, speedThresholdForMoving, notMovingNumberOfFramesThresholdForSleep, maxDistBetweenTwoPointsInsideSleepingPeriod=-1, specifiedStartTime=0, distanceTravelledRollingMedianFilter=0, videoPixelSize=-1, videoFPS=-1):
  
  videoNamelist    = []
  dfFinalAllVideos = pd.DataFrame()
  currentNbFrames  = 0
  startSleepPeriod = -1
  
  if ',' in vidName:
    videoNamelist = vidName.split(',')
  else:
    videoNamelist = [vidName]
  
  for idx, videoName in enumerate(videoNamelist):

    try:
      with openResultsFile(os.path.join(pathToZZoutput, videoName), 'r') as results:
        dataRef = createSuperStructFromH5(results)
    except ValueError:  # h5 results not found, assume a results folder exists
      pathToVideo = os.path.join(pathToZZoutput, videoName, "results_" + videoName + ".txt")
      with open(pathToVideo) as ff:
        dataRef = json.load(ff)

    if videoPixelSize == -1:
      if "videoPixelSize" in dataRef:
        videoPixelSize = dataRef["videoPixelSize"]
      else:
        print("You should add the parameter 'videoPixelSize' in your configuration file")
        videoPixelSize = 1
    
    if videoFPS == -1:
      if "videoFPS" in dataRef:
        videoFPS = dataRef["videoFPS"]
      else:
        print("You should add the parameter 'videoFPS' in your configuration file")
        videoFPS = 1
    
    if idx == 0 and type(specifiedStartTime) == str and len(specifiedStartTime) > 1:
      t = datetime.datetime.strptime(specifiedStartTime, "%H:%M:%S")
      currentNbFrames = (t.hour * 60 * 60 + t.minute * 60 + t.second) * videoFPS
    
    dfFinal = pd.DataFrame()
    totalFrames  = len(dataRef['wellPoissMouv'][0][0][0]["HeadX"])
    times = [time.strftime('%H:%M:%S', time.gmtime(second)) for second in [(i + currentNbFrames) / videoFPS for i in range(0, totalFrames + 1)]]
    currentNbFrames = currentNbFrames + totalFrames + 1
    df = pd.DataFrame(data = np.transpose([times]), columns = ['HourMinuteSecond'])
    dfFinal = pd.concat([dfFinal, df], axis=1)
    
    for wellNumber in range(len(dataRef['wellPoissMouv'])):
      
      headPosition = np.array([dataRef['wellPoissMouv'][wellNumber][0][0]["HeadX"], dataRef['wellPoissMouv'][wellNumber][0][0]["HeadY"]])
      displacementVector = videoPixelSize * np.diff(headPosition)
      squaredDisplacement = displacementVector * displacementVector
      distance = np.sqrt(squaredDisplacement[0] + squaredDisplacement[1])
      # distance = np.append(distance, [0])
      distance = np.insert(distance, 0, 0)
      distance = np.insert(distance, 0, 0)
      headPosition = np.concatenate((np.zeros((2,1)), headPosition), axis=1)
      if distanceTravelledRollingMedianFilter:
        distance2 = np.convolve(distance, np.ones(distanceTravelledRollingMedianFilter), 'same') / distanceTravelledRollingMedianFilter
        distance2[:distanceTravelledRollingMedianFilter-1] = distance[:distanceTravelledRollingMedianFilter-1]
        distance2[-distanceTravelledRollingMedianFilter+1:] = distance[-distanceTravelledRollingMedianFilter+1:]
        distance = distance2
      
      speed = distance * videoFPS
      
      moving = speed > speedThresholdForMoving
      
      sleep = np.array([False for i in range(len(moving))])
      countNotMovingFrames = 0
      for i in range(len(sleep)):
        if moving[i]:
          countNotMovingFrames = 0
          startSleepPeriod     = -1
        else:
          if maxDistBetweenTwoPointsInsideSleepingPeriod == -1 or startSleepPeriod == -1 or (videoPixelSize*math.sqrt((headPosition[0][startSleepPeriod] - headPosition[0][i])**2 + (headPosition[1][startSleepPeriod] - headPosition[1][i])**2)) < maxDistBetweenTwoPointsInsideSleepingPeriod:
            countNotMovingFrames = countNotMovingFrames + 1
            if countNotMovingFrames == 1:
              startSleepPeriod = i
            if countNotMovingFrames >= notMovingNumberOfFramesThresholdForSleep:
              sleep[i] = True
            if countNotMovingFrames == notMovingNumberOfFramesThresholdForSleep:
              countNotMovingFrames2 = countNotMovingFrames - 1
              j = i - 1
              while j >= 0 and countNotMovingFrames2 > 0:
                sleep[j] = True
                countNotMovingFrames2 = countNotMovingFrames2 - 1
                j = j - 1
          else:
            countNotMovingFrames = 0
            startSleepPeriod     = -1
      
      df = pd.DataFrame(data = np.transpose(np.append(np.append([speed], [moving], axis=0), [sleep], axis=0)), columns = ['speed_' + str(wellNumber), 'moving_' + str(wellNumber), 'sleep_' + str(wellNumber)])
      dfFinal = pd.concat([dfFinal, df], axis=1)

    resultsFolder = os.path.join(pathToZZoutput, videoName)
    if not os.path.exists(resultsFolder):
      os.makedirs(resultsFolder)
    dfFinal.to_excel(os.path.join(resultsFolder, "sleepVsMoving_" + videoName + ".xlsx"))
    dfFinalAllVideos = pd.concat([dfFinalAllVideos, dfFinal], axis=0)
    
  if len(videoNamelist) > 1:
    name1 = videoNamelist[0]
    name2 = videoNamelist[1]
    match = SequenceMatcher(None, name1, name2).find_longest_match(0, len(name1), 0, len(name2))
    commonSubstring = name1[match.a: match.a + match.size]
    concatExcelFileName = commonSubstring + '_'.join([name.replace(commonSubstring, '') for name in videoNamelist])
    dfFinalAllVideos.to_excel(os.path.join(pathToZZoutput, "sleepVsMoving_" + concatExcelFileName + ".xlsx"))

def firstSleepingTimeAfterSpecifiedTime(pathToZZoutput, vidName, specifiedTime, wellNumber):
  
  if ',' in vidName:
    videoNamelist = vidName.split(',')
    name1 = videoNamelist[0]
    name2 = videoNamelist[1]
    match = SequenceMatcher(None, name1, name2).find_longest_match(0, len(name1), 0, len(name2))
    commonSubstring = name1[match.a: match.a + match.size]
    concatExcelFileName = commonSubstring + '_'.join([name.replace(commonSubstring, '') for name in videoNamelist])
    df = pd.read_excel(os.path.join(pathToZZoutput, "sleepVsMoving_" + concatExcelFileName + ".xlsx"))
  else:
    concatExcelFileName = vidName
    df = pd.read_excel(os.path.join(os.path.join(pathToZZoutput, vidName), "sleepVsMoving_" + vidName + ".xlsx"))
  
  dfTimeMinusSpecifiedTime = (df['HourMinuteSecond'].apply(datetime.datetime.strptime, args=("%H:%M:%S",)) - datetime.datetime.strptime(specifiedTime, "%H:%M:%S")).apply(pd.Timedelta.total_seconds).apply(abs)
  indexStart = dfTimeMinusSpecifiedTime.argmin()
  dfLength = len(df)
  
  if int(wellNumber) == -1:
    allWellNumbers = [str(i) for i in range(0, int(len(df.columns) / 3))]
    dfResult = pd.DataFrame(index=[i for i in range(0, int(len(df.columns) / 3))], columns=['firstSleepTime', 'delayBeforeFirstSleepTimeInSeconds'])
  else:
    allWellNumbers = [wellNumber]
    dfResult = pd.DataFrame(index=[int(wellNumber)], columns=['firstSleepTime', 'delayBeforeFirstSleepTimeInSeconds'])
  
  for wellNum in allWellNumbers:
    currentIndex = indexStart
    while (currentIndex != dfLength) and (df['sleep_' + wellNum][currentIndex] == 0):
      currentIndex = currentIndex + 1
    if currentIndex == dfLength:
      print("For the well number", wellNum, ", couldn't find any time point after", specifiedTime, "for which the fish was sleeping")
    else:
      print("For the well number:", wellNum, "and the specified time:", specifiedTime, ".")
      print("The first time after the specified time when the fish starts sleeping is at", df['HourMinuteSecond'][currentIndex], "(which corresponds to frame number:", currentIndex, ").")
      print("In other words, the fish starts sleeping", dfTimeMinusSpecifiedTime[currentIndex], "seconds after the specified time (or", currentIndex - indexStart, "frames after the specified time) (or", time.strftime('%H:%M:%S', time.gmtime(dfTimeMinusSpecifiedTime[currentIndex])), "Hours:Minutes:Seconds after the specified time)")
      dfResult['firstSleepTime'][int(wellNum)]                     = df['HourMinuteSecond'][currentIndex]
      dfResult['delayBeforeFirstSleepTimeInSeconds'][int(wellNum)] = dfTimeMinusSpecifiedTime[currentIndex]
  
  dfResult.to_excel(os.path.join(pathToZZoutput, "firstSleepTimeAfter_" + specifiedTime.replace(':', '') + "_" + concatExcelFileName + "_" + wellNumber + ".xlsx"))

def numberOfSleepingAndMovingTimesInTimeRange(pathToZZoutput, vidName, specifiedStartTime, specifiedEndTime, wellNumber):
  
  if ',' in vidName:
    videoNamelist = vidName.split(',')
    name1 = videoNamelist[0]
    name2 = videoNamelist[1]
    match = SequenceMatcher(None, name1, name2).find_longest_match(0, len(name1), 0, len(name2))
    commonSubstring = name1[match.a: match.a + match.size]
    concatExcelFileName = commonSubstring + '_'.join([name.replace(commonSubstring, '') for name in videoNamelist])
    df = pd.read_excel(os.path.join(pathToZZoutput, "sleepVsMoving_" + concatExcelFileName + ".xlsx"))
  else:
    concatExcelFileName = vidName
    df = pd.read_excel(os.path.join(os.path.join(pathToZZoutput, vidName), "sleepVsMoving_" + vidName + ".xlsx"))
  
  dfTimeMinusSpecifiedStartTime = (df['HourMinuteSecond'].apply(datetime.datetime.strptime, args=("%H:%M:%S",)) - datetime.datetime.strptime(specifiedStartTime, "%H:%M:%S")).apply(pd.Timedelta.total_seconds).apply(abs)
  indexStart = dfTimeMinusSpecifiedStartTime.argmin()
  
  dfTimeMinusSpecifiedEndTime = (df['HourMinuteSecond'].apply(datetime.datetime.strptime, args=("%H:%M:%S",)) - datetime.datetime.strptime(specifiedEndTime, "%H:%M:%S")).apply(pd.Timedelta.total_seconds).apply(abs)
  indexEnd = dfTimeMinusSpecifiedEndTime.argmin()
  
  dfLength = len(df)
  
  if int(wellNumber) == -1:
    allWellNumbers = [str(i) for i in range(0, int(len(df.columns) / 3))]
    dfResult = pd.DataFrame(index=[i for i in range(0, int(len(df.columns) / 3))], columns=['nbSleepFrames', 'nbMovingFrames'])
  else:
    allWellNumbers = [wellNumber]
    dfResult = pd.DataFrame(index=[int(wellNumber)], columns=['nbSleepFrames', 'nbMovingFrames'])
  
  print("Between time", specifiedStartTime, "(frame number", indexStart, "), and time:", specifiedEndTime, "(frame number", indexEnd,"):")
  
  for wellNum in allWellNumbers:
  
    print("For well number", wellNum, ":")
    
    nbSleepFrames  = np.sum(df['sleep_'  + wellNum][indexStart:indexEnd])
    nbMovingFrames = np.sum(df['moving_' + wellNum][indexStart:indexEnd])
    
    dfResult['nbSleepFrames'][int(wellNum)]  = nbSleepFrames
    dfResult['nbMovingFrames'][int(wellNum)] = nbMovingFrames
    
    print(nbSleepFrames, "sleeping Frames")
    print(nbMovingFrames, "moving frames\n")
  
  dfResult.to_excel(os.path.join(pathToZZoutput, "nbSleepAndMoveFrames_" + concatExcelFileName + "_" + specifiedStartTime.replace(':', '') + "_" + specifiedEndTime.replace(':', '') + "_" + wellNumber + ".xlsx"))


def numberOfSleepBoutsInTimeRange(pathToZZoutput, vidName, minSleepLenghtDurationThreshold, wellNumber='-1', specifiedStartTime=-1, specifiedEndTime=-1):
  
  if ',' in vidName:
    videoNamelist = vidName.split(',')
    name1 = videoNamelist[0]
    name2 = videoNamelist[1]
    match = SequenceMatcher(None, name1, name2).find_longest_match(0, len(name1), 0, len(name2))
    commonSubstring = name1[match.a: match.a + match.size]
    concatExcelFileName = commonSubstring + '_'.join([name.replace(commonSubstring, '') for name in videoNamelist])
    df = pd.read_excel(os.path.join(pathToZZoutput, "sleepVsMoving_" + concatExcelFileName + ".xlsx"))
  else:
    concatExcelFileName = vidName
    df = pd.read_excel(os.path.join(os.path.join(pathToZZoutput, vidName), "sleepVsMoving_" + vidName + ".xlsx"))
 
  if type(specifiedStartTime) != int:
    dfTimeMinusSpecifiedStartTime = (df['HourMinuteSecond'].apply(datetime.datetime.strptime, args=("%H:%M:%S",)) - datetime.datetime.strptime(specifiedStartTime, "%H:%M:%S")).apply(pd.Timedelta.total_seconds).apply(abs)
    indexStart = dfTimeMinusSpecifiedStartTime.argmin()
    dfTimeMinusSpecifiedEndTime = (df['HourMinuteSecond'].apply(datetime.datetime.strptime, args=("%H:%M:%S",)) - datetime.datetime.strptime(specifiedEndTime, "%H:%M:%S")).apply(pd.Timedelta.total_seconds).apply(abs)
    indexEnd = dfTimeMinusSpecifiedEndTime.argmin()
  else:
    indexStart = 0
    indexEnd   = len(df)
  
  dfLength = len(df)
  
  if int(wellNumber) == -1:
    allWellNumbers = [str(i) for i in range(0, int(len(df.columns) / 3))]
    dfResult = pd.DataFrame(index=[i for i in range(0, int(len(df.columns) / 3))], columns=['nbSleepBouts'])
  else:
    allWellNumbers = [wellNumber]
    dfResult = pd.DataFrame(index=[int(wellNumber)], columns=['nbSleepBouts'])
  
  if type(specifiedStartTime) != int:
    print("Between time", specifiedStartTime, "(frame number", indexStart, "), and time:", specifiedEndTime, "(frame number", indexEnd,"):")
  
  for wellNum in allWellNumbers:
  
    print("For well number", wellNum, ":")
    
    currentlySleeping = False
    curSleepLength = 0
    countNbSleepBoutsOverDurationThreshold = 0
    for i in range(indexStart, indexEnd):
      if df['sleep_'  + wellNum][i] == 1:
        currentlySleeping = True
        curSleepLength = curSleepLength + 1
      else:
        if currentlySleeping:
          if curSleepLength >= minSleepLenghtDurationThreshold:
            countNbSleepBoutsOverDurationThreshold = countNbSleepBoutsOverDurationThreshold + 1
        currentlySleeping = False
        curSleepLength = 0

    if curSleepLength >= minSleepLenghtDurationThreshold:
      countNbSleepBoutsOverDurationThreshold = countNbSleepBoutsOverDurationThreshold + 1    

    dfResult['nbSleepBouts'][int(wellNum)]  = countNbSleepBoutsOverDurationThreshold
    
    print(str(countNbSleepBoutsOverDurationThreshold), "sleeping Frames\n")
  
  if type(specifiedStartTime) == int:
    specifiedStartTime = ''
    specifiedEndTime   = ''
  
  dfResult.to_excel(os.path.join(pathToZZoutput, "nbSleepBouts_" + concatExcelFileName + "_" + specifiedStartTime.replace(':', '') + "_" + specifiedEndTime.replace(':', '') + "_" + wellNumber + ".xlsx"))



from ._openResultsFile import openResultsFile


def getConfigurationFileUsed(videoName: str) -> dict:
  with openResultsFile(videoName, 'r') as results:
    if 'configurationFileUsed' not in results:
      raise ValueError(f'configuration file not found in {results.filename}')
    return dict(results['configurationFileUsed'].attrs)



import os
import matplotlib.pyplot as plt
import math
import pickle

def distBetweenAngles(angle1, angle2):
  diffAngle = angle2 - angle1
  if diffAngle < -math.pi:
    diffAngle = diffAngle + 2 * math.pi
  if diffAngle > math.pi:
    diffAngle = diffAngle - 2 * math.pi
  return diffAngle

def computeEyesHeadingPlot(superStruct, hyperparameters, videoNameWithTimestamp):
  
  outputPath = os.path.join(hyperparameters["outputFolder"], videoNameWithTimestamp)
  if not os.path.exists(outputPath):
    os.makedirs(outputPath)
  
  for wellId in range(0, len(superStruct["wellPoissMouv"])):
    for animalId in range(0, len(superStruct["wellPoissMouv"][wellId])):
      for boutId in range(0, len(superStruct["wellPoissMouv"][wellId][animalId])):
      
        leftEyeAngle   = superStruct["wellPoissMouv"][wellId][animalId][boutId]["leftEyeAngle"]
        rightEyeAngle  = superStruct["wellPoissMouv"][wellId][animalId][boutId]["rightEyeAngle"]
        heading        = superStruct["wellPoissMouv"][wellId][animalId][boutId]["Heading"]
        leftEyeArea    = superStruct["wellPoissMouv"][wellId][animalId][boutId]["leftEyeArea"]
        rightEyeArea   = superStruct["wellPoissMouv"][wellId][animalId][boutId]["rightEyeArea"]
        leftDiffAngle  = [distBetweenAngles(angle, heading[idx]) for idx, angle in enumerate(leftEyeAngle)]
        rightDiffAngle = [distBetweenAngles(angle, heading[idx]) for idx, angle in enumerate(rightEyeAngle)]
        
        if True:
          fig, tabAx = plt.subplots(2, 1, figsize=(22.9, 8.8))
          tabAx[0].plot([idx for idx, val in enumerate(leftDiffAngle)], leftDiffAngle)
          tabAx[0].set_title('Head direction angle - Left eye direction angle')
          tabAx[1].plot([idx for idx, val in enumerate(rightDiffAngle)], rightDiffAngle)
          tabAx[1].set_title('Head direction angle - Right eye direction angle')   
        else:
          fig, tabAx = plt.subplots(4, 1, figsize=(22.9, 8.8))
          tabAx[0].plot([idx for idx, val in enumerate(leftDiffAngle)], leftDiffAngle)
          tabAx[0].set_title('Head direction angle - Left eye direction angle')
          tabAx[1].plot([idx for idx, val in enumerate(leftEyeArea)], leftEyeArea)
          tabAx[1].set_title('Left eye area')
          tabAx[2].plot([idx for idx, val in enumerate(rightDiffAngle)], rightDiffAngle)
          tabAx[2].set_title('Head direction angle - Right eye direction angle')
          tabAx[3].plot([idx for idx, val in enumerate(rightEyeArea)], rightEyeArea)
          tabAx[3].set_title('Right eye area')
        
        plt.savefig(os.path.join(outputPath, hyperparameters["videoName"] + "_eyeDiffAngle_" + str(wellId) + '_' + str(animalId) + '_' + str(boutId) + '.png'))
        
        dataToSave = {"leftDiffAngle": leftDiffAngle, "rightDiffAngle": rightDiffAngle, "leftEyeArea": leftEyeArea, "rightEyeArea": rightEyeArea}
        with open(os.path.join(outputPath, hyperparameters["videoName"] + "_eyeDiffAnglePickleData_" + str(wellId) + '_' + str(animalId) + '_' + str(boutId) + '.pickle'), "wb") as f:
          pickle.dump(dataToSave, f)
        


from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.tailTrackFindNextPoint import tailTrackFindNextPoint
from zebrazoom.code.extractParameters import calculateTailAngle
import numpy as np
import math
import cv2

def _calculateAngle(xStart, yStart, xEnd, yEnd):
  vx = xEnd - xStart
  vy = yEnd - yStart
  if vx == 0:
    if vy > 0:
      lastFirstTheta = math.pi/2
    else:
      lastFirstTheta = (3*math.pi)/2
  else:
    lastFirstTheta = np.arctan(abs(vy/vx))
    if (vx < 0) and (vy >= 0):
      lastFirstTheta = math.pi - lastFirstTheta
    elif (vx < 0) and (vy <= 0):
      lastFirstTheta = lastFirstTheta + math.pi
    elif (vx > 0) and (vy <= 0):
      lastFirstTheta = 2*math.pi - lastFirstTheta
  return lastFirstTheta

def _getCurvature(points):
  tailX = points[0]
  tailY = points[1]
  l = len(tailX)
  if l > 2:
    curvature = np.zeros(l-2)
    for ii in range(1, l-1):
      angleBef = _calculateAngle(tailX[ii-1], tailY[ii-1], tailX[ii], tailY[ii])
      angleAft = _calculateAngle(tailX[ii],   tailY[ii],   tailX[ii+1], tailY[ii+1])
      curvature[ii-1] = calculateTailAngle(angleBef, angleAft)
    return curvature
  else:
    return []

def tailTrackTry4InitialPerpendicularDirections(headPosition, frameROI, points, lastFirstTheta, maxDepth, steps, nbList, hyperparameters, debug, lenX, lenY):
  (points1, lastFirstTheta1, medianPixTotList1) = tailTrackFindNextPoint(0, headPosition, frameROI, points, lastFirstTheta, maxDepth, steps, nbList,  hyperparameters, debug, lenX, lenY)
  (points2, lastFirstTheta2, medianPixTotList2) = tailTrackFindNextPoint(0, headPosition, frameROI, points, (lastFirstTheta + (math.pi/2)) % (2 * math.pi), maxDepth, steps, nbList,  hyperparameters, debug, lenX, lenY)
  (points3, lastFirstTheta3, medianPixTotList3) = tailTrackFindNextPoint(0, headPosition, frameROI, points, (lastFirstTheta + math.pi) % (2 * math.pi), maxDepth, steps, nbList,  hyperparameters, debug, lenX, lenY)
  (points4, lastFirstTheta4, medianPixTotList4) = tailTrackFindNextPoint(0, headPosition, frameROI, points, (lastFirstTheta + (3/2)*math.pi) % (2 * math.pi), maxDepth, steps, nbList,  hyperparameters, debug, lenX, lenY)
  #
  listOfPoints           = [points1, points2, points3, points4]
  listOfLastFirstTheta   = [lastFirstTheta1, lastFirstTheta2, lastFirstTheta3, lastFirstTheta4]
  listOfMedianPixTotList = [medianPixTotList1, medianPixTotList2, medianPixTotList3, medianPixTotList4]
  listOfCurvatures       = [_getCurvature(point) for point in listOfPoints]
  listOfCurvaturesMeans  = [np.mean(curvature) if len(curvature) else 10000000000 for curvature in listOfCurvatures]
  #
  if False:
    selectedOptionPix  = np.argmin(listOfMedianPixTotList)
    selectedOptionCurv = np.argmin(listOfCurvaturesMeans)
    selectedOption     = selectedOptionCurv
  else:
    argsortOptionPix  = np.argsort(listOfMedianPixTotList)
    argsortOptionCurv = np.argsort(listOfCurvaturesMeans)
    scores = np.ones(len(listOfMedianPixTotList))
    for i in range(len(scores)):
      scores[argsortOptionPix[i]]  += i
      scores[argsortOptionCurv[i]] += i
    selectedOptionTemp = np.argmin(scores)
    possibleInd = (scores < scores[selectedOptionTemp] * 1.5)
    possibleInds = []
    selectedOption = 0
    curMin = 1000000000000000000000000000000000
    for i in range(len(scores)):
      if possibleInd[i] and listOfMedianPixTotList[i] < curMin:
        selectedOption = i
        curMin = listOfMedianPixTotList[i]
    # print("listOfMedianPixTotList:", listOfMedianPixTotList)
    # print("listOfCurvaturesMeans:", listOfCurvaturesMeans)
    # print("scores:", scores)
    # print("possibleInd:", possibleInd)
    # print("selectedOption:", selectedOption)
  #
  points           = listOfPoints[selectedOption]
  lastFirstTheta   = listOfLastFirstTheta[selectedOption]
  medianPixTotList = listOfMedianPixTotList[selectedOption]
  # print("frameNumber:", frameNumber, ";selectedOption:", selectedOption, "; lastFirstTheta:", lastFirstTheta)
  
  return (points, lastFirstTheta, medianPixTotList)



from zebrazoom.code.findWells import findWells
from zebrazoom.code.getHyperparameters import getHyperparameters

import csv

import os

from zebrazoom.code.tracking import get_default_tracking_method


def getTailExtremityFirstFrame(pathToVideo, videoName, videoExt, configFile, argv):
  videoNameWithoutExt = videoName
  videoName = videoName + '.' + videoExt
  
  videoPath = os.path.join(pathToVideo, videoName)

  # Getting hyperparameters
  [hyperparameters, config] = getHyperparameters(configFile, videoName, videoPath, argv)
  
  frameNumber = hyperparameters["firstFrame"]
  wellNumber = 0
  if hyperparameters["oneWellManuallyChosenTopLeft"]:
    wellPositions = findWells(os.path.join(pathToVideo, videoName), hyperparameters)
  else:
    wellPositions = [{"topLeftX":0, "topLeftY":0, "lengthX": hyperparameters["videoWidth"], "lengthY": hyperparameters["videoHeight"]}]
  tracking = get_default_tracking_method()(videoPath, wellPositions, hyperparameters)
  [frame, thresh1] = tracking.headEmbededFrame(frameNumber, wellNumber)

  inputsFolder = os.path.join(hyperparameters['outputFolder'], '.ZebraZoomVideoInputs', videoNameWithoutExt)
  if not os.path.exists(inputsFolder):
    os.makedirs(inputsFolder)
  frame = tracking.getAccentuateFrameForManualPointSelect(frame)
  if hyperparameters["findHeadPositionByUserInput"]:
    with open(os.path.join(inputsFolder, f'{videoNameWithoutExt}HP.csv'), mode='w') as f:
      writer = csv.writer(f, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
      writer.writerow(tracking.findHeadPositionByUserInput(frame, frameNumber, wellNumber))

  with open(os.path.join(inputsFolder, f'{videoNameWithoutExt}.csv'), mode='w') as f:
      writer = csv.writer(f, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
      writer.writerow(tracking.findTailTipByUserInput(frame, frameNumber, wellNumber))



import math

import numpy as np

from zebrazoom.code.extractParameters import calculateAngle, calculateTailAngle, smoothAllTailAngles
from zebrazoom.code.getHyperparameters import getHyperparametersSimple


def calculateAndStoreTailAngleHeatmap(results, dataGroup, boutsGroup):
  hyperparameters = getHyperparametersSimple(dict(results['configurationFileUsed'].attrs))
  firstFrame = results.attrs['firstFrame']
  lastFrame = results.attrs['lastFrame']
  pointsToTakeIntoAccountStart = 9 - int(hyperparameters["tailAnglesHeatMapNbPointsToTakeIntoAccount"])
  allAngles = None
  data = np.empty(lastFrame - firstFrame + 1, dtype=[(f'Pos{idx}', float) for idx in range(1, hyperparameters['nbTailPoints'] - 1)])
  data[:] = np.nan
  for bout in boutsGroup:
    boutGroup = boutsGroup[bout]
    start = boutGroup.attrs['BoutStart'] - firstFrame
    end = boutGroup.attrs['BoutEnd'] - firstFrame + 1

    if 'allTailAnglesSmoothed' not in boutsGroup[bout]:  # calculate all tail angles
      if allAngles is None:
        TailX_VideoReferential = np.column_stack([dataGroup['HeadPos']['X']] + [dataGroup['TailPosX'][col] for col in dataGroup['TailPosX'].attrs['columns']])
        TailY_VideoReferential = np.column_stack([dataGroup['HeadPos']['Y']] + [dataGroup['TailPosY'][col] for col in dataGroup['TailPosY'].attrs['columns']])
        Heading = np.array(dataGroup['Heading'])
        frames, points = TailX_VideoReferential.shape
        allAngles = np.zeros((frames, points))
        for i in range(frames):
          head = np.array([TailX_VideoReferential[i, 0], TailY_VideoReferential[i, 0]])
          for j in range(points):
            allAngles[i][j] = calculateTailAngle(calculateAngle(head, np.array([TailX_VideoReferential[i, j], TailY_VideoReferential[i, j]])), (Heading[i] + math.pi) % (2 * math.pi))

      allTailAngles, allTailAnglesSmoothed = smoothAllTailAngles(allAngles, hyperparameters, start, end - 1)
      boutGroup.create_dataset('allTailAngles', data=allTailAngles)
      boutGroup.create_dataset('allTailAnglesSmoothed', data=allTailAnglesSmoothed)
      tailAngles = list(allTailAnglesSmoothed)[pointsToTakeIntoAccountStart:]
    else:
      tailAngles = list(boutsGroup[bout]['allTailAnglesSmoothed'])[pointsToTakeIntoAccountStart:]

    for idx, tailAngle in enumerate(tailAngles):  # calculate tail angle heatmap
      if end - start == len(tailAngle):
        data[f'Pos{idx + 1}'][start:end] = [t * (180 / math.pi) for t in tailAngle]

  dataset = dataGroup.create_dataset('tailAngleHeatmap', data=data)
  dataset.attrs['columns'] = data.dtype.names
  return data



import math

import cv2
import numpy as np


class ComputeHeadingMixin:
  def _computeHeading(self, thresh1, x, y, headSize):
    if headSize == -1:
      print("Setting headSize to 25 instead of -1, this may be a problem in some cases")
      headSize = 25 # This was introduced to fix bug when config file creation for center of mass only tracking 16/11/22

    videoWidth  = self._hyperparameters["videoWidth"]
    videoHeight = self._hyperparameters["videoHeight"]

    x = x + int(headSize)
    y = y + int(headSize)

    paddedImage = np.zeros((len(thresh1) + 2 * int(headSize), len(thresh1[0]) + 2 * int(headSize)))
    paddedImage[:, :] = 255
    paddedImage[int(headSize):len(thresh1)+int(headSize), int(headSize):len(thresh1[0])+int(headSize)] = thresh1
    thresh1 = paddedImage

    thresh1 = thresh1.astype(np.uint8)

    ymin  = y - headSize
    ymax  = y + headSize
    xmin  = x - headSize
    xmax  = x + headSize

    img = thresh1[int(ymin):int(ymax), int(xmin):int(xmax)]

    img[0,:] = 255
    img[len(img)-1,:] = 255
    img[:,0] = 255
    img[:,len(img[0])-1] = 255

    y, x = np.nonzero(img)
    x = x - np.mean(x)
    y = y - np.mean(y)
    coords = np.vstack([x, y])
    cov = np.cov(coords)
    evals, evecs = np.linalg.eig(cov)
    sort_indices = np.argsort(evals)[::-1]
    x_v1, y_v1 = evecs[:, sort_indices[0]]  # Eigenvector with largest eigenvalue
    x_v2, y_v2 = evecs[:, sort_indices[1]]
    scale = 20
    theta = self._calculateAngle(0, 0, x_v1, y_v1)
    theta = (theta - math.pi/2) % (2 * math.pi)

    if self._hyperparameters["debugHeadingCalculation"]:
      img2 = img.copy()
      img2 = cv2.cvtColor(img2, cv2.COLOR_GRAY2BGR)
      cv2.line(img2, (int(len(img2[0])/2), int(len(img2)/2)), (int(len(img[0])/2 + 20 * math.cos(theta)), int(len(img)/2 + 20 * math.sin(theta))), (255,0,0), 1)
      self._debugFrame(img2, title='imgForHeadingCalculation')

    return theta

  def _calculateHeading(self, x, y, i, thresh1, thresh2, takeTheHeadClosestToTheCenter, previousFrameHeading=0):
    nbList = self._hyperparameters["nbList"]
    expDecreaseFactor = self._hyperparameters["expDecreaseFactor"]
    headSize = self._hyperparameters["headSize"]
    debugTracking = self._hyperparameters["debugTracking"]
    headEmbeded = self._hyperparameters["headEmbeded"]

    if self._hyperparameters["debugHeadingCalculation"]:
      self._debugFrame(thresh1, title='thresh1')
      self._debugFrame(thresh2, title='thresh2')

    cx = 0
    cy = 0
    cxNew = 0
    cyNew = 0

    thresh1[0,:]                 = 255
    thresh1[len(thresh1)-1,:]    = 255
    thresh1[:,0]                 = 255
    thresh1[:,len(thresh1[0])-1] = 255

    contours, hierarchy = cv2.findContours(thresh1,cv2.RETR_TREE,cv2.CHAIN_APPROX_SIMPLE)
    for contour in contours:
      dist = cv2.pointPolygonTest(contour, (float(x), float(y)), True)
      if dist >= 0:
        M = cv2.moments(contour)
        if M['m00']:
          cx = int(M['m10']/M['m00'])
          cy = int(M['m01']/M['m00'])
          bodyContour = contour
        else:
          cx = 0
          cy = 0
        if self._hyperparameters["debugHeadingCalculation"]:
          print("area:", area)
          print("dist, cx, cy:", dist, cx, cy)
          print("x, y:", x, y)

    if self._hyperparameters["debugHeadingCalculation"]:
      print("x, y, cx, cy:", x, y, cx, cy)

    heading = self._computeHeading(thresh2, x, y, headSize)
    if self._hyperparameters["debugHeadingCalculation"]:
      print("previousFrameHeading:", previousFrameHeading)
      print("math.sqrt((x - cx)**2 + (y - cy)**2):", math.sqrt((x - cx)**2 + (y - cy)**2))
    if math.sqrt((x - cx)**2 + (y - cy)**2) > 1:
      lastFirstTheta = self._calculateAngle(x, y, cx, cy)
    else:
      lastFirstTheta = (previousFrameHeading + math.pi) % (2 * math.pi)

    headingApproximate = lastFirstTheta
    headingPreciseOpt1 = heading
    headingPreciseOpt2 = (heading + math.pi) % (2*math.pi)
    diffAngle1 = self._distBetweenThetas(headingApproximate, headingPreciseOpt1)
    diffAngle2 = self._distBetweenThetas(headingApproximate, headingPreciseOpt2)
    if (diffAngle1 > diffAngle2):
      heading = headingPreciseOpt1
    else:
      heading = headingPreciseOpt2

    if self._hyperparameters["debugHeadingCalculation"]:
      print("headingApproximate:", headingApproximate)
      print("headingPreciseOpt1:", headingPreciseOpt1)
      print("headingPreciseOpt2:", headingPreciseOpt2)
      print("diffAngle1:", diffAngle1)
      print("diffAngle2:", diffAngle2)
      print("heading:", heading)

    return [heading, lastFirstTheta]

  def _calculateHeadingSimple(self, x, y, thresh2):
    headSize = self._hyperparameters["headSize"]
    heading = self._computeHeading(thresh2, x, y, headSize)
    return heading



from ._openResultsFile import openResultsFile


def setFPSandPixelSize(videoName: str, videoFPS: float, videoPixelSize: float) -> None:

  with openResultsFile(videoName, 'a') as results:
    if videoFPS == results.attrs.get('videoFPS') and videoPixelSize == results.attrs.get('videoPixelSize'):
      return
    results.attrs['videoFPS']       = float(videoFPS)
    results.attrs['videoPixelSize'] = float(videoPixelSize)

    # if kinematic parameters were calculated, remove them since they are no longer valid
    for wellIdx in range(len(results['wellPositions'])):
      wellGroup = results[f'dataForWell{wellIdx}']
      for animal in wellGroup:
        animalGroup = wellGroup[animal]
        if 'kinematicParametersPerBout' in animalGroup:
          del animalGroup['kinematicParametersPerBout']
        if 'listOfBouts' in animalGroup:
          for bout in animalGroup['listOfBouts']:
            boutGroup = animalGroup[f'listOfBouts/{bout}']
            if 'additionalKinematicParametersPerBout' in boutGroup:
              del boutGroup['additionalKinematicParametersPerBout']
    if 'frameStepForDistanceCalculation' in results.attrs:
      del results.attrs['frameStepForDistanceCalculation']



import os
import math
import shutil
import matplotlib.pyplot as plt
import pickle
import numpy as np
import pandas as pd
import json
import seaborn as sns

from zebrazoom.dataAnalysis.dataanalysis import sortGenotypes

def populationComparaison(nameOfFile, resFolder, globParam, conditions, genotypes, outputFolder, medianPerWellFirstForEachKinematicParameter = 0, plotOutliersAndMean = True, saveDataPlottedInJson = 0, medianPerGenotypeFirstForEachKinematicParameter=0, numberOfBoutsPerSecond=0):

  outputFolderResult = os.path.join(outputFolder, nameOfFile)
  if not(os.path.exists(outputFolderResult)):
    os.mkdir(outputFolderResult)
    
  if medianPerWellFirstForEachKinematicParameter:
    outputFolderResult = os.path.join(outputFolderResult, 'medianPerWellFirst')
  else:
    if medianPerGenotypeFirstForEachKinematicParameter:
      outputFolderResult = os.path.join(outputFolderResult, 'medianPerGenotypeFirst')
    else:
      outputFolderResult = os.path.join(outputFolderResult, 'allBoutsMixed')
  
  if plotOutliersAndMean: # This is a little bit of a hack because the creation of folders relies on the order in which the populationComparaison function is called (relative to the plotOutliersAndMean parameter)
    if os.path.exists(outputFolderResult): 
      shutil.rmtree(outputFolderResult)
    while True:
      try:
        os.mkdir(outputFolderResult)
        break
      except:
        print("waiting to create folder:", outputFolderResult)
    outputFolderCharts = outputFolderResult
  else:
    outputFolderCharts = os.path.join(outputFolderResult, 'noMeanAndOutliersPlotted')
    os.makedirs(outputFolderCharts)
  
  dataPlotted = {}
  
  if os.path.exists(os.path.join(resFolder, nameOfFile + '.pkl')):
    infile = open(os.path.join(resFolder, nameOfFile + '.pkl'),'rb')
  else:
    infile = open(os.path.join(resFolder, nameOfFile),'rb') # This is just to insure compatibility with previous versions, should remove this line in the future
  dfParam = pickle.load(infile)
  infile.close()
  
  columnsForRawDataExport = ['Trial_ID', 'Well_ID', 'NumBout', 'BoutStart', 'BoutEnd', 'Condition', 'Genotype', 'videoDuration'] + globParam
  
  if medianPerWellFirstForEachKinematicParameter:
    dfKinematicValues = dfParam[columnsForRawDataExport]
    dfKinematicValues = dfKinematicValues.astype({param: float for param in globParam + ['videoDuration']})
    dfKinematicValues = dfKinematicValues.groupby(['Trial_ID', 'Well_ID']).median(numeric_only=True)
    dfCondGeno = dfParam[['Trial_ID', 'Well_ID', 'Condition', 'Genotype']]
    dfCondGeno = dfCondGeno.groupby(['Trial_ID', 'Well_ID']).first()
    dfCount = dfParam[['Trial_ID', 'Well_ID']].copy()
    dfCount['Bout Counts'] = dfParam['Bout Duration (s)'] # IS THERE A PROBLEM HERE???
    dfCount = dfCount.groupby(['Trial_ID', 'Well_ID']).count()
    dfCount['Bout Rate (bouts / s)'] = dfCount['Bout Counts'] / dfKinematicValues['videoDuration']
    dfTotalTimeMove = dfParam[['Trial_ID', 'Well_ID', 'Bout Duration (s)']].copy()
    dfTotalTimeMove = dfTotalTimeMove.groupby(['Trial_ID', 'Well_ID']).sum()
    dfKinematicValues['percentTimeSpentSwimming'] = (dfTotalTimeMove['Bout Duration (s)'] / dfKinematicValues['videoDuration']) * 100
    dfParam = pd.concat([dfCondGeno, dfKinematicValues], axis=1)
    dfParam = pd.concat([dfParam, dfCount], axis=1)
    globParam = globParam + ['percentTimeSpentSwimming', 'Bout Counts', 'Bout Rate (bouts / s)']
    dfParamForExcelExport = dfParam.copy()
  elif medianPerGenotypeFirstForEachKinematicParameter:
    dfKinematicValues = dfParam[columnsForRawDataExport]
    dfKinematicValues = dfKinematicValues.astype({param: float for param in globParam})
    dfKinematicValues = dfKinematicValues.groupby(['Genotype', 'Well_ID']).median(numeric_only=True)
    dfCond = dfParam[['Genotype', 'Well_ID', 'Condition']]
    dfCond = dfCond.groupby(['Genotype', 'Well_ID']).first()
    dfCount = dfParam[['Genotype', 'Well_ID']].copy()
    dfCount['Bout Counts'] = [0 for i in range(len(dfCount['Genotype']))]
    dfCount = dfCount.groupby(['Genotype', 'Well_ID']).count()
    dfCount['Bout Rate (bouts / s)'] = dfCount['Bout Counts'] / dfKinematicValues['videoDuration']
    dfTotalTimeMove = dfParam[['Trial_ID', 'Well_ID', 'Bout Duration (s)']].copy()
    dfTotalTimeMove = dfTotalTimeMove.groupby(['Trial_ID', 'Well_ID']).sum()
    dfKinematicValues['percentTimeSpentSwimming'] = (dfTotalTimeMove['Bout Duration (s)'] / dfKinematicValues['videoDuration']) * 100
    dfParam = pd.concat([dfCond, dfKinematicValues], axis=1)
    dfParam = pd.concat([dfParam, dfCount], axis=1)
    globParam = globParam + ['percentTimeSpentSwimming', 'Bout Counts', 'Bout Rate (bouts / s)']
    dfParamForExcelExport = dfParam.copy()
  else:
    columnsForRawDataExport2 = columnsForRawDataExport.copy()
    columnsForRawDataExport2.insert(2, "Animal_ID")
    dfParamForExcelExport = dfParam[columnsForRawDataExport2].copy()
    dfParam  = dfParam[columnsForRawDataExport]
  
  if not os.path.exists(os.path.join(outputFolderResult, 'globalParametersInsideCategories.xlsx')):
    dfParamForExcelExport.to_excel(os.path.join(outputFolderResult, 'globalParametersInsideCategories.xlsx'))
    dfParamForExcelExport.to_csv(os.path.join(outputFolderResult, 'globalParametersInsideCategories.csv'), index=False)

  genotypes = dfParam["Genotype"].unique().tolist()
  palette = dict(zip(sortGenotypes(genotypes), sns.color_palette(n_colors=len(genotypes))))
  nbGraphs = int(len(globParam)/6) if len(globParam) % 6 == 0 else int(len(globParam)/6) + 1
  color = ['b', 'r', 'c', 'm', 'y', 'k']
  for i in range(nbGraphs):
    globParamForPlot = [globParam[elem] for elem in range(6*i, min(6*(i+1), len(globParam)))]
    nbLines   = int(math.sqrt(len(globParamForPlot)))
    nbColumns = math.ceil(len(globParamForPlot) / nbLines)
    fig, tabAx = plt.subplots(nbLines, nbColumns, figsize=(22.9, 8.8))
    fig.tight_layout(pad=4.0)
    for idx, parameter in enumerate(globParamForPlot):
      print("plotting parameter:", parameter)
      
      if True and not(medianPerGenotypeFirstForEachKinematicParameter):
        
        tabToPlot = 0
        if nbLines == 1:
          if nbColumns == 1:
            tabToPlot = tabAx
          else:
            tabToPlot = tabAx[idx%nbColumns]
        else:
          tabToPlot = tabAx[int(idx/nbColumns), idx%nbColumns]
        
        b = sns.boxplot(ax=tabToPlot, data=dfParam, x="Condition", y=parameter, hue="Genotype", showmeans=plotOutliersAndMean, showfliers=plotOutliersAndMean, palette=palette, hue_order=palette.keys())
        # if plotOutliersAndMean and (medianPerWellFirstForEachKinematicParameter or medianPerGenotypeFirstForEachKinematicParameter): # or len(concatenatedValues[0]) < 100):
          # sns.stripplot(ax=tabToPlot, data=dfParam, x="Condition", y=parameter, hue="Genotype", color=".25")
        b.set_ylabel('', fontsize=0)
        b.set_xlabel('', fontsize=0)
        b.axes.set_title(parameter,fontsize=20)
        
      # else: # Old Method (TO REMOVE!!!), and for the medianPerGenotypeFirstForEachKinematicParameter option
      
        # concatenatedValues = []
        # labels = []
        # if not(medianPerGenotypeFirstForEachKinematicParameter):
          # for condition in conditions:
            # for genotype in genotypes:
              # indicesCondition = dfParam.index[dfParam['Condition'] == condition].tolist()
              # indicesGenotype  = dfParam.index[dfParam['Genotype']  == genotype].tolist()
              # indices = [ind for ind in indicesCondition if ind in indicesGenotype]
              # values  = dfParam.loc[indices, parameter].values
              # concatenatedValues.append(values)
              # labels.append(str(condition) + '\n' + str(genotype))
        # else:
          # for condition in conditions:
            # indicesCondition = dfParam.index[dfParam['Condition'] == condition].tolist()
            # values  = dfParam.loc[indicesCondition, parameter].values
            # concatenatedValues.append(values)
            # labels.append(str(condition))
        
        # concatenatedValuesWithoutNans = []
        # for toConcat in concatenatedValues:
          # concatenatedValuesWithoutNans.append(np.array([x for x in toConcat if not(math.isnan(x))]))
        # concatenatedValues = concatenatedValuesWithoutNans
        
        # if saveDataPlottedInJson:
          # dataPlotted[parameter] = {}
          # for idx2, label in enumerate(labels):
            # dataPlotted[parameter][label] = concatenatedValues[idx2].tolist()
        
        # if nbLines == 1:
          # if nbColumns == 1:
            # tabAx.set_title(parameter)
            # tabAx.boxplot(concatenatedValues, showmeans=plotOutliersAndMean, showfliers=plotOutliersAndMean)
            # if plotOutliersAndMean and (medianPerWellFirstForEachKinematicParameter or medianPerGenotypeFirstForEachKinematicParameter or len(concatenatedValues[0]) < 100):
              # for idx2, values in enumerate(concatenatedValues):
                # tabAx.plot(np.random.normal(idx2+1, 0.005*len(concatenatedValues), size=len(values)), values, 'b.', alpha=0.3, c=color[idx2] if idx2 < len(color) else 'b')
            # tabAx.set_xticklabels(labels)
          # else:
            # tabAx[idx%nbColumns].set_title(parameter)
            # tabAx[idx%nbColumns].boxplot(concatenatedValues, showmeans=plotOutliersAndMean, showfliers=plotOutliersAndMean)
            # if plotOutliersAndMean and (medianPerWellFirstForEachKinematicParameter or medianPerGenotypeFirstForEachKinematicParameter or len(concatenatedValues[0]) < 100):
              # for idx2, values in enumerate(concatenatedValues):
                # tabAx[idx%nbColumns].plot(np.random.normal(idx2+1, 0.005*len(concatenatedValues), size=len(values)), values, 'b.', alpha=0.3, c=color[idx2] if idx2 < len(color) else 'b')
            # tabAx[idx%nbColumns].set_xticklabels(labels)
        # else:
          # tabAx[int(idx/nbColumns), idx%nbColumns].set_title(parameter)
          # tabAx[int(idx/nbColumns), idx%nbColumns].boxplot(concatenatedValues, showmeans=plotOutliersAndMean, showfliers=plotOutliersAndMean)
          # if plotOutliersAndMean and (medianPerWellFirstForEachKinematicParameter or medianPerGenotypeFirstForEachKinematicParameter or len(concatenatedValues[0]) < 100):
            # for idx2, values in enumerate(concatenatedValues):
              # tabAx[int(idx/nbColumns), idx%nbColumns].plot(np.random.normal(idx2+1, 0.005*len(concatenatedValues), size=len(values)), values, 'b.', alpha=0.3, c=color[idx2] if idx2 < len(color) else 'b')
          # tabAx[int(idx/nbColumns), idx%nbColumns].set_xticklabels(labels)
    
    plt.savefig(os.path.join(outputFolderCharts, 'globalParametersInsideCategories_' + str(i+1) + '.png'))
    plt.close(fig)
  
  if saveDataPlottedInJson:
    outputFile = open(os.path.join(outputFolderCharts, 'dataPlotted.txt'), 'w')
    outputFile.write(json.dumps(dataPlotted))
    outputFile.close()
  return globParam, dfParam



import numpy as np

from ._openResultsFile import openResultsFile


def getDataPerBout(videoName: str, numWell: int, numAnimal: int, numBout: int, parameterName: str) -> np.array:
  with openResultsFile(videoName, 'r') as results:
    boutPath = f'dataForWell{numWell}/dataForAnimal{numAnimal}/listOfBouts/bout{numBout}'
    perFramePath = f'dataForWell{numWell}/dataForAnimal{numAnimal}/dataPerFrame'
    if boutPath not in results:
      raise ValueError(f"bout {numBout} for animal {numAnimal} in well {numWell} doesn't exist")
    boutGroup = results[boutPath]
    firstFrame = results.attrs['firstFrame']
    start = boutGroup.attrs['BoutStart'] - firstFrame
    end = boutGroup.attrs['BoutEnd'] - firstFrame + 1
    dataGroup = results[perFramePath]
    if parameterName not in dataGroup:
      allParameters = ', '.join(dataGroup.keys())
      raise ValueError(f'Parameter "{parameterName}" not found in results. Available parameters: {allParameters}')
    dataset = dataGroup[parameterName]
    return np.array(dataset[start:end]) if 'columns' not in dataset.attrs else np.column_stack([dataset[col][start:end] for col in dataset.attrs['columns']])



from ._base import BaseTrackingMethod, register_tracking_method, get_tracking_method, get_default_tracking_method
from ._baseZebraZoom import BaseZebraZoomTrackingMethod
from ._getBackground import getBackground
from . import tracking
from . import fasterMultiprocessing
from . import fasterMultiprocessing2



import numpy as np

from ._calculateAndStoreCurvature import calculateAndStoreCurvature
from ._openResultsFile import openResultsFile


def getCurvaturePerTimeInterval(videoName: str, numWell: int, numAnimal: int, startTimeInSeconds: float, endTimeInSeconds: float) -> [np.array, np.array, np.array]:
  if startTimeInSeconds >= endTimeInSeconds:
    raise ValueError('end time must be larger than start time')
  with openResultsFile(videoName, 'r+') as results:
    firstFrame = results.attrs['firstFrame']
    firstFrameInSeconds = firstFrame / results.attrs['videoFPS']
    lastFrameInSeconds = results.attrs['lastFrame'] / results.attrs['videoFPS']
    if startTimeInSeconds < firstFrameInSeconds or endTimeInSeconds > lastFrameInSeconds:
      raise ValueError(f'Tracking was performed from {firstFrameInSeconds}s to {lastFrameInSeconds}s, start and end times must be within this interval.')
    if 'videoFPS' not in results.attrs:
      raise ValueError(f'videoFPS not found in the results, cannot convert seconds to frames')
    if 'videoPixelSize' not in results.attrs:
      raise ValueError(f'videoPixelSize not found in the results, cannot convert seconds to frames')
    intervalStart = int(startTimeInSeconds * results.attrs['videoFPS']) - firstFrame
    intervalEnd   = int(endTimeInSeconds * results.attrs['videoFPS']) - firstFrame
    dataGroup = results[f'dataForWell{numWell}/dataForAnimal{numAnimal}/dataPerFrame']
    
    if 'curvature' in dataGroup:
      curvature = np.array([dataGroup['curvature'][column][intervalStart:intervalEnd] for column in dataGroup['curvature'].attrs['columns']])
    else:
      print(f'calculating and storing curvature for all frames for well {numWell}, animal {numAnimal}')
      curvature = np.array([data[intervalStart:intervalEnd] for data in calculateAndStoreCurvature(results, dataGroup)])
    
    # Getting x time values for each curvature point
    boutStart = int(startTimeInSeconds * results.attrs['videoFPS'])
    boutEnd   = int(endTimeInSeconds   * results.attrs['videoFPS'])
    xTimeValues = curvature.copy()
    for i in range(0, len(xTimeValues[0])):
      xTimeValues[:, i] = int(100*((i + boutStart)/results.attrs['videoFPS']))/100
    
    # Getting y distance along the tail values for each curvature point
    perFramePath = f'dataForWell{numWell}/dataForAnimal{numAnimal}/dataPerFrame'
    start = boutStart - firstFrame
    end   = boutEnd   - firstFrame
    tailX = np.transpose(np.concatenate((np.array([results[perFramePath+"/HeadPos"]['X'][start:end]]).reshape((-1, 1)), results[perFramePath+"/TailPosX"][start:end].view((float, len(results[perFramePath+"/TailPosX"][start:end].dtype.names)))), axis=1))
    tailY = np.transpose(np.concatenate((np.array([results[perFramePath+"/HeadPos"]['Y'][start:end]]).reshape((-1, 1)), results[perFramePath+"/TailPosY"][start:end].view((float, len(results[perFramePath+"/TailPosX"][start:end].dtype.names)))), axis=1))
    nbLinesTail = len(tailX)
    yDistanceAlongTheTail = curvature.copy()
    yDistanceAlongTheTail[:, :] = np.sqrt(np.square(tailX[2:nbLinesTail, :] - tailX[1:nbLinesTail-1, :]) + np.square(tailY[2:nbLinesTail, :] - tailY[1:nbLinesTail-1, :])) * results.attrs['videoPixelSize']
    for i in range(len(yDistanceAlongTheTail)-2, -1, -1):
      yDistanceAlongTheTail[i, :] += yDistanceAlongTheTail[i+1, :]

    return [curvature, xTimeValues, yDistanceAlongTheTail]



import math

import cv2
import numpy as np
from numpy import linspace
from scipy.interpolate import UnivariateSpline

from ._tailTrackingBase import TailTrackingBase


class TailTrackingBlobDescentMixin(TailTrackingBase):
  @staticmethod
  def __dist2(v, w):
    return (v["x"] - w["x"])**2 + (v["y"] - w["y"])**2

  def __distToSegmentSquared(self, p, v, w):
    l2 = self.__dist2(v, w)
    if l2 == 0:
      return self.__dist2(p, v)
    t = ((p["x"] - v["x"]) * (w["x"] - v["x"]) + (p["y"] - v["y"]) * (w["y"] - v["y"])) / l2
    t = max(0, min(1, t))
    return self.__dist2(p, {"x": v["x"] + t * (w["x"] - v["x"]), "y": v["y"] + t * (w["y"] - v["y"]) })

  def __distToSegment(self, p, v, w):
    return math.sqrt(self.__distToSegmentSquared(p, v, w))

  def __checkNewPointNotRedundant(self, points, x, y):
    if len(points[0]) > 3:
      dist = (points[0] - x)**2 + (points[1] - y)**2
      min1 = np.argmin(dist)
      minDist = dist[min1]

      p = {"x": x, "y": y}
      for i in range(0,len(points[0])-1):
        v = {"x": points[0][i],   "y": points[1][i]   }
        w = {"x": points[0][i+1], "y": points[1][i+1] }
        d = self.__distToSegment(p, v, w)
        if d < minDist:
          minDist = d

      if minDist <= 1:
        return 0
      else:
        return 1
    else:
      return 1

  def __recenterPointAlongOrthogonalTailAxis(self, x, y, theta, frame, thresh1):
    l = [[x,y]]

    thresh1lenX = len(thresh1[0]) - 1
    thresh1lenY = len(thresh1) - 1

    pixVal = 1
    k1 = 1
    while pixVal < 150 and k1 < 100:
      xNew = self._assignValueIfBetweenRange(int(x + k1 * (math.cos(theta)) ), 0, thresh1lenX)
      yNew = self._assignValueIfBetweenRange(int(y + k1 * (math.sin(theta)) ), 0, thresh1lenY)
      pixVal = thresh1[yNew][xNew]
      if pixVal < 150:
        k1 = k1 + 1
        cv2.circle(frame, (xNew, yNew), 1, (0,0,0),   -1)
        point = [xNew, yNew]
        l.append(point)

    pixVal = 1
    k2 = 1
    while pixVal < 150 and k2 < 100:
      xNew = self._assignValueIfBetweenRange(int(x - k2 * (math.cos(theta)) ), 0, thresh1lenX)
      yNew = self._assignValueIfBetweenRange(int(y - k2 * (math.sin(theta)) ), 0, thresh1lenY)
      pixVal = thresh1[yNew][xNew]
      if pixVal < 150:
        k2 = k2 + 1
        cv2.circle(frame, (xNew, yNew), 1, (0,0,0),   -1)
        point = [xNew, yNew]
        l = [point] + l

    n = len(l)
    xNew = self._assignValueIfBetweenRange(int((l[0][0] + l[n-1][0]) / 2), 0, thresh1lenX)
    yNew = self._assignValueIfBetweenRange(int((l[0][1] + l[n-1][1]) / 2), 0, thresh1lenY)

    return [xNew, yNew]

  def __findNextPoints(self, x, y, thresh1, frame, depth, points, lastTheta, debugAdv, nbList, thetaDiffAccept, expDecreaseFactor, step, debugTracking):
    thresh1lenX = len(thresh1[0]) - 1
    thresh1lenY = len(thresh1) - 1

    maxThetaCorrected = -1

    if debugAdv:
      self._debugFrame(frame, title='Frame')

    if (depth < 25): #15):
      maxTheta = 0
      ktotMax = 0
      k1Max = 0
      k2Max = 0

      l = [i*(math.pi/nbList) for i in range(0,2*nbList) if abs(i*(math.pi/nbList)-lastTheta) < thetaDiffAccept]

      for thetaB in l:
        theta = thetaB % math.pi

        # Find furtherest point away from "old" "current" point still inside blob along the theta angle in one direction
        pixVal = 1
        k1 = 1
        while pixVal < 150 and k1 < 100:
          xNew = self._assignValueIfBetweenRange(int(x + k1 * (math.cos(theta))), 0, thresh1lenX)
          yNew = self._assignValueIfBetweenRange(int(y + k1 * (math.sin(theta))), 0, thresh1lenY)
          pixVal = thresh1[yNew][xNew]
          k1 = k1 + 1
        k1 = k1 - 2

        # Find furtherest point away from "old" "current" point still inside blob along the theta angle in the other direction
        pixVal = 1
        k2 = 1
        while pixVal < 150 and k2 < 100:
          xNew = self._assignValueIfBetweenRange(int(x - k2 * (math.cos(theta)) ), 0, thresh1lenX)
          yNew = self._assignValueIfBetweenRange(int(y - k2 * (math.sin(theta)) ), 0, thresh1lenY)
          pixVal = thresh1[yNew][xNew]
          k2 = k2 + 1
        k2 = k2 - 2
        ktot = k1 + k2

        # Calculate "score" for this theta angle based on the lenght of the segment fitted inside the blob
        ktot = 0
        for i in range(0,k1):
          ktot = ktot + math.exp(-i*expDecreaseFactor)
        for i in range(0,k2):
          ktot = ktot + math.exp(-i*expDecreaseFactor)

        # Keeps that theta angle as maximum if appropriate
        if (ktot > ktotMax):
          ktotMax = ktot
          k1Max = k1
          k2Max = k2
          maxTheta = theta

      # Calculates the two new possible points on both side of the local tail angle derivate
      if k1Max > step:
        x1 = self._assignValueIfBetweenRange(int(x + step  * (math.cos(maxTheta)) ), 0, thresh1lenX)
        y1 = self._assignValueIfBetweenRange(int(y + step  * (math.sin(maxTheta)) ), 0, thresh1lenY)
      else:
        x1 = self._assignValueIfBetweenRange(int(x + k1Max  * (math.cos(maxTheta)) ), 0, thresh1lenX)
        y1 = self._assignValueIfBetweenRange(int(y + k1Max  * (math.sin(maxTheta)) )  , 0, thresh1lenY)
      if k2Max > step:
        x2 = self._assignValueIfBetweenRange(int(x - step  * (math.cos(maxTheta)) ), 0, thresh1lenX)
        y2 = self._assignValueIfBetweenRange(int(y - step  * (math.sin(maxTheta)) ), 0, thresh1lenY)
      else:
        x2 = self._assignValueIfBetweenRange(int(x - k2Max  * (math.cos(maxTheta)) ), 0, thresh1lenX)
        y2 = self._assignValueIfBetweenRange(int(y - k2Max  * (math.sin(maxTheta)) ), 0, thresh1lenY)

      # Sanity check
      if debugTracking:
        if thresh1[y2][x2] > 150:
          print("oups white zone forw",k2Max)
        if thresh1[y1][x1] > 150:
          print("oups white zone back",k1Max)

      if (depth == 0):
        predictedTetha = maxTheta
      else:
        predictedTetha = maxTheta + ((maxTheta - lastTheta + 2*math.pi) % (2*math.pi))
      [x1, y1] = self.__recenterPointAlongOrthogonalTailAxis(x1, y1, predictedTetha+(math.pi)/2, frame, thresh1)
      [x2, y2] = self.__recenterPointAlongOrthogonalTailAxis(x2, y2, predictedTetha+(math.pi)/2, frame, thresh1)

      x1 = self._assignValueIfBetweenRange(x1, 0, thresh1lenX)
      y1 = self._assignValueIfBetweenRange(y1, 0, thresh1lenY)
      x2 = self._assignValueIfBetweenRange(x2, 0, thresh1lenX)
      y2 = self._assignValueIfBetweenRange(y2, 0, thresh1lenY)

      cv2.circle(frame, (x, y), 1, (255,0,255),   -1)

      # Calculates distance between new and old local tail angle
      diffAngle1 = self._distBetweenThetas(lastTheta, maxTheta )
      diffAngle2 = self._distBetweenThetas(lastTheta, (maxTheta + math.pi) )

      # Calculates distance between new and old point
      distSubsquentPoints1 = (x1 - x)**2 + (y1 - y)**2
      distSubsquentPoints2 = (x2 - x)**2 + (y2 - y)**2

      # If the distance between new and old point larger than 0, appends points and launch search for new point
      if (diffAngle1 > diffAngle2):
        if debugTracking:
          if thresh1[y2][x2] > 150:
            print("oups white zone")
          if debugAdv:
            print("diffAngle2: ",diffAngle2," ; k2Max: ",k2Max," distSubsquentPoints2:",distSubsquentPoints2)
        cv2.circle(frame, (x2, y2), 1, (0,255,0),   -1)
        check = self.__checkNewPointNotRedundant(points, x2, y2)
        if check:
          points = self._appendPoint(x2, y2, points)
          if distSubsquentPoints2 > 0:
            maxThetaCorrected = maxTheta + math.pi
            newTheta = self._calculateAngle(x,y,x2,y2)
            (points,nop) = self.__findNextPoints(x2,y2,thresh1,frame,depth+1,points,newTheta,debugAdv,nbList,thetaDiffAccept,expDecreaseFactor,step,debugTracking)
      else:
        if debugTracking:
          if thresh1[y1][x1] > 150:
            print("oups white zone")
          if debugAdv:
            print("diffAngle1: ",diffAngle1," ; k1Max: ",k1Max," distSubsquentPoints1:",distSubsquentPoints1)
        cv2.circle(frame, (x1, y1), 1, (0,255,0),   -1)
        check = self.__checkNewPointNotRedundant(points, x1, y1)
        if check:
          points = self._appendPoint(x1, y1, points)
          if distSubsquentPoints1 > 0:
            maxThetaCorrected = maxTheta
            newTheta = self._calculateAngle(x,y,x1,y1)
            (points,nop) = self.__findNextPoints(x1,y1,thresh1,frame,depth+1,points,newTheta,debugAdv,nbList,thetaDiffAccept,expDecreaseFactor,step,debugTracking)

    return (points,maxThetaCorrected)

  @staticmethod
  def __smoothTail(points, nbTailPoints):
    y = points[0]
    x = linspace(0, 1, len(y))

    if len(x) > 3:

      s = UnivariateSpline(x, y, s=10)
      xs = linspace(0, 1, nbTailPoints)
      newX = s(xs)

      y = points[1]
      x = linspace(0, 1, len(y))
      s = UnivariateSpline(x, y, s=10)
      xs = linspace(0, 1, nbTailPoints)
      newY = s(xs)
    else:
      newX = x
      newY = y

    return [newX, newY]

  def _tailTrackingBlobDescent(self, headPosition, i, thresh1, frame, lastFirstTheta, debugAdv, thetaDiffAccept):
    step = self._hyperparameters["step"]
    nbList = self._hyperparameters["nbList"]
    expDecreaseFactor = self._hyperparameters["expDecreaseFactor"]
    firstFrame = self._hyperparameters["firstFrame"]
    lastFrame = self._hyperparameters["lastFrame"]
    minArea = self._hyperparameters["minArea"]
    maxArea = self._hyperparameters["maxArea"]
    headSize = self._hyperparameters["headSize"]
    debugTracking = self._hyperparameters["debugTracking"]
    headEmbeded = self._hyperparameters["headEmbeded"]
    x, y = headPosition

    output = np.zeros((1, self._nbTailPoints, 2))

    possiblePoints = []
    for ste in step:
      points = np.zeros((2, 0))
      (points, lastFirstTheta2) = self.__findNextPoints(x,y,thresh1,frame,0,points,lastFirstTheta,debugAdv,nbList,thetaDiffAccept,expDecreaseFactor,ste,debugTracking)
      possiblePoints.append(points)

    nbPossiblePoints = len(possiblePoints)
    minDist = 1000
    pBest = 0
    if nbPossiblePoints == 1:
      points = possiblePoints[0]
    else:
      for p1 in range(0, nbPossiblePoints):
        for p2 in range(0, nbPossiblePoints):
          if p2 > p1:
            points1 = possiblePoints[p1]
            n1 = len(points1[0]) - 1
            x1 = int(points1[0][n1])
            y1 = int(points1[1][n1])

            points2 = possiblePoints[p2]
            n2 = len(points2[0]) - 1
            x2 = int(points2[0][n2])
            y2 = int(points2[1][n2])

            dist = math.sqrt((x1-x2)**2 + (y1-y2)**2)
            if dist < minDist:
              minDist = dist
              pBest = p1
      points = possiblePoints[pBest]

    points = np.insert(points, 0, headPosition, axis=1)
    points = self.__smoothTail(points, self._nbTailPoints)

    # for idx, x in enumerate(points[0]):
      # output[i-firstFrame][idx][0] = x
      # output[i-firstFrame][idx][1] = points[1][idx]

    for idx, x in enumerate(points[0]):
      output[0][idx][0] = x
      output[0][idx][1] = points[1][idx]

    return output



from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.detectMovementWithRawVideoInsideTracking import detectMovementWithRawVideoInsideTracking
from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.trackTail import trackTail
from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.animalDetectionBackToWholeImage import animalDetectionBackToWholeImage
from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.removeAnimalJustTrackedFromFrame import removeAnimalJustTrackedFromFrame
# import zebrazoom.code.tracking
import numpy as np
import time
import math
import cv2

def backgroundSubtractionOnlyOnROIs(self, frame, k):
  
  # Color to grey scale transformation
  if type(frame[0][0]) == np.ndarray:
    frame = frame[:,:,0]
  
  # Bout detection
  if self._hyperparameters["detectMovementWithRawVideoInsideTracking"]:
    prevFrame = detectMovementWithRawVideoInsideTracking(self, k, frame)
    if ("detectMovementCompareWithTheFuture" in self._hyperparameters) and self._hyperparameters["detectMovementCompareWithTheFuture"]:
      frame = prevFrame
  
  if ("paramGaussianBlurForHeadPosition" in self._hyperparameters) and self._hyperparameters["paramGaussianBlurForHeadPosition"]:
    frameGaussianBlurForHeadPosition = frame.copy()
  
  # Going through each well/arena/tank and applying tracking method on it
  t1 = time.time()
  for wellNumber in self._listOfWellsOnWhichToRunTheTracking:
    
    if self._hyperparameters["detectMovementWithRawVideoInsideTracking"] == 0 or k <= 2 or np.sum([self._auDessusPerAnimalIdList[wellNumber][i][k] for i in range(0, self._hyperparameters["nbAnimalsPerWell"])]):
      
      animalNotTracked = np.zeros((self._hyperparameters["nbAnimalsPerWell"]))
      
      for animalId in range(0, self._hyperparameters["nbAnimalsPerWell"]):
        
        if self._hyperparameters["detectMovementWithRawVideoInsideTracking"] == 0 or self._auDessusPerAnimalIdList[wellNumber][animalId][k] or k <= 2:
          
          # Retrieving ROI coordinates and selecting ROI
          roiXStart = self._wellPositions[wellNumber]['topLeftX'] + int(self._trackingDataPerWell[wellNumber][animalId][k-1][0][0] - self._hyperparameters["backgroundSubtractionOnROIhalfDiameter"])
          roiYStart = self._wellPositions[wellNumber]['topLeftY'] + int(self._trackingDataPerWell[wellNumber][animalId][k-1][0][1] - self._hyperparameters["backgroundSubtractionOnROIhalfDiameter"])
          roiXEnd   = roiXStart + 2 * self._hyperparameters["backgroundSubtractionOnROIhalfDiameter"]
          roiYEnd   = roiYStart + 2 * self._hyperparameters["backgroundSubtractionOnROIhalfDiameter"]
          if roiXStart < 0:
            roiXStart = 0
            roiXEnd   = 2 * self._hyperparameters["backgroundSubtractionOnROIhalfDiameter"]
          if roiYStart < 0:
            roiYStart = 0
            roiYEnd   = 2 * self._hyperparameters["backgroundSubtractionOnROIhalfDiameter"]
          if roiXEnd >= len(frame[0]):
            roiXEnd   = len(frame[0]) - 1
            roiXStart = len(frame[0]) - 2 * self._hyperparameters["backgroundSubtractionOnROIhalfDiameter"]
          if roiYEnd >= len(frame):
            roiYEnd   = len(frame) - 1
            roiYStart = len(frame) - 2 * self._hyperparameters["backgroundSubtractionOnROIhalfDiameter"]
          frameROI = frame[roiYStart:roiYEnd, roiXStart:roiXEnd].copy()
          if ("paramGaussianBlurForHeadPosition" in self._hyperparameters) and self._hyperparameters["paramGaussianBlurForHeadPosition"]:
            frameGaussianBlurForHeadPositionROI = frameGaussianBlurForHeadPosition[roiYStart:roiYEnd, roiXStart:roiXEnd].copy()
          
          # Subtracting background of image
          if not("noBackgroundSubtraction" in self._hyperparameters) or not(self._hyperparameters["noBackgroundSubtraction"]):
            backgroundROI = self._background[roiYStart:roiYEnd, roiXStart:roiXEnd]
            frameROI = 255 - np.where(backgroundROI >= frameROI, backgroundROI - frameROI, 0).astype(np.uint8)
            if ("paramGaussianBlurForHeadPosition" in self._hyperparameters) and self._hyperparameters["paramGaussianBlurForHeadPosition"]:
              frameGaussianBlurForHeadPositionROI = 255 - np.where(backgroundROI >= frameGaussianBlurForHeadPositionROI, backgroundROI - frameGaussianBlurForHeadPositionROI, 0).astype(np.uint8)

          # Applying gaussian filter
          paramGaussianBlur = self._hyperparameters["paramGaussianBlur"]
          frameROI = cv2.GaussianBlur(frameROI, (paramGaussianBlur, paramGaussianBlur), 0)
          if ("paramGaussianBlurForHeadPosition" in self._hyperparameters) and self._hyperparameters["paramGaussianBlurForHeadPosition"]:
            frameGaussianBlurForHeadPositionROI = cv2.GaussianBlur(frameGaussianBlurForHeadPositionROI, (self._hyperparameters["paramGaussianBlurForHeadPosition"], self._hyperparameters["paramGaussianBlurForHeadPosition"]), 0)
          
          # Head position tracking
          if ("paramGaussianBlurForHeadPosition" in self._hyperparameters) and self._hyperparameters["paramGaussianBlurForHeadPosition"]:
            (minVal, maxVal, headPosition, maxLoc) = cv2.minMaxLoc(frameGaussianBlurForHeadPositionROI)
          else:
            (minVal, maxVal, headPosition, maxLoc) = cv2.minMaxLoc(frameROI)
          if minVal >= self._hyperparameters["minimumHeadPixelValue"]:
            self._trackingDataPerWell[wellNumber][animalId][k] = self._trackingDataPerWell[wellNumber][animalId][k-1]
            animalNotTracked[animalId] = 1
          else:
            if self._hyperparameters["trackTail"]:
              # Tail tracking
              a, self._lastFirstTheta[wellNumber] = trackTail(self, frameROI, headPosition, self._hyperparameters, wellNumber, k, self._lastFirstTheta[wellNumber])
              if len(a):
                a[0, :, 0] += roiXStart - self._wellPositions[wellNumber]['topLeftX']
                a[0, :, 1] += roiYStart - self._wellPositions[wellNumber]['topLeftY']
                self._trackingDataPerWell[wellNumber][animalId][k][:len(a[0])] = a
              else:
                self._trackingDataPerWell[wellNumber][animalId][k] = self._trackingDataPerWell[wellNumber][animalId][k-1]
                animalNotTracked[animalId] = 1
            else:
              self._trackingDataPerWell[wellNumber][animalId][k][0][0] = headPosition[0] + roiXStart - self._wellPositions[wellNumber]['topLeftX']
              self._trackingDataPerWell[wellNumber][animalId][k][0][1] = headPosition[1] + roiYStart - self._wellPositions[wellNumber]['topLeftY']
        else:
          if k > 0:
            animalNotTracked[animalId] = 1
            self._trackingDataPerWell[wellNumber][animalId][k] = self._trackingDataPerWell[wellNumber][animalId][k-1]
        
        # 'Removing' animal just tracked
        if self._hyperparameters["nbAnimalsPerWell"] > 1:
          frame = removeAnimalJustTrackedFromFrame(self, frame, wellNumber, animalId, k)
          if ("paramGaussianBlurForHeadPosition" in self._hyperparameters) and self._hyperparameters["paramGaussianBlurForHeadPosition"]:
            frameGaussianBlurForHeadPosition = removeAnimalJustTrackedFromFrame(self, frameGaussianBlurForHeadPosition, wellNumber, animalId, k)
      
      if "animalDetectionBackToWholeImage" in self._hyperparameters and self._hyperparameters["animalDetectionBackToWholeImage"] and np.sum(animalNotTracked):
        animalDetectionBackToWholeImage(self, k, frame, wellNumber, animalNotTracked)
            
      # Id invertion if necessary
      if self._hyperparameters["nbAnimalsPerWell"] > 1 and False:
        # NEED TO IMPROVE THIS IN THE FUTURE!!!
        for animalId in range(0, self._hyperparameters["nbAnimalsPerWell"]-1):
          dist_animalIdCurFrame_animal0PrevFrame = math.sqrt((self._trackingDataPerWell[wellNumber][animalId][k][0][0] - self._trackingDataPerWell[wellNumber][0][k-1][0][0])**2     +     (self._trackingDataPerWell[wellNumber][animalId][k][0][1] - self._trackingDataPerWell[wellNumber][0][k-1][0][1])**2)
          dist_animalIdCurFrame_animal1PrevFrame = math.sqrt((self._trackingDataPerWell[wellNumber][animalId][k][0][0] - self._trackingDataPerWell[wellNumber][1][k-1][0][0])**2     +     (self._trackingDataPerWell[wellNumber][animalId][k][0][1] - self._trackingDataPerWell[wellNumber][1][k-1][0][1])**2)
          if dist_animalIdCurFrame_animal1PrevFrame < dist_animalIdCurFrame_animal0PrevFrame:
            temp = self._trackingDataPerWell[wellNumber][0][k].copy()
            self._trackingDataPerWell[wellNumber][0][k] = self._trackingDataPerWell[wellNumber][1][k]
            self._trackingDataPerWell[wellNumber][1][k] = temp
      
      # Checking for superimposed tracking, removing them if necessary
      if ("checkAndRemoveSuperimposedTrackingIfPresent_minDistance" in self._hyperparameters) and (self._hyperparameters["checkAndRemoveSuperimposedTrackingIfPresent_minDistance"] != 0):
        for animalId1 in range(0, self._hyperparameters["nbAnimalsPerWell"]-1):
          for animalId2 in range(animalId1 + 1, self._hyperparameters["nbAnimalsPerWell"]):
            min_distBetweenTwoPointsOfAnimals = 10000000000000000000000
            for p1 in range(len(self._trackingDataPerWell[wellNumber][animalId1][k])):
              for p2 in range(len(self._trackingDataPerWell[wellNumber][animalId2][k])):
                distBetweenTwoPointsOfAnimals = math.sqrt((self._trackingDataPerWell[wellNumber][animalId1][k][p1][0] - self._trackingDataPerWell[wellNumber][animalId2][k][p2][0])**2   +   (self._trackingDataPerWell[wellNumber][animalId1][k][p1][1] - self._trackingDataPerWell[wellNumber][animalId2][k][p2][1])**2)
                if distBetweenTwoPointsOfAnimals < min_distBetweenTwoPointsOfAnimals:
                  min_distBetweenTwoPointsOfAnimals = distBetweenTwoPointsOfAnimals
            if min_distBetweenTwoPointsOfAnimals < self._hyperparameters["checkAndRemoveSuperimposedTrackingIfPresent_minDistance"]:
              # randomly choosing to remove animalId1, in the future should better choose which one to remove
              for p1 in range(len(self._trackingDataPerWell[wellNumber][animalId1][k])):
                self._trackingDataPerWell[wellNumber][animalId1][k][p1][0] = 0
                self._trackingDataPerWell[wellNumber][animalId1][k][p1][1] = 0
    else:  
      for animalId in range(0, self._hyperparameters["nbAnimalsPerWell"]):
        self._trackingDataPerWell[wellNumber][animalId][k] = self._trackingDataPerWell[wellNumber][animalId][k-1]


from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.utilities import calculateAngle, distBetweenThetas
from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.detectMovementWithTrackedDataAfterTracking import detectMovementWithTrackedDataAfterTracking
from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.detectMovementWithRawVideoInsideTracking import detectMovementWithRawVideoInsideTracking
from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.getListOfWellsOnWhichToRunTheTracking import getListOfWellsOnWhichToRunTheTracking
from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.backgroundSubtractionOnWholeImage import backgroundSubtractionOnWholeImage
from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.backgroundSubtractionOnlyOnROIs import backgroundSubtractionOnlyOnROIs
import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
from zebrazoom.code.extractParameters import extractParameters
import zebrazoom.code.util as util
import zebrazoom.code.tracking
import numpy as np
import queue
import math
import time
import cv2

from ..._updateBackgroundAtInterval import UpdateBackgroundAtIntervalMixin


class Tracking(zebrazoom.code.tracking.BaseTrackingMethod, UpdateBackgroundAtIntervalMixin):
  
  def __init__(self, videoPath, wellPositions, hyperparameters):
    self._videoPath = videoPath
    self._wellPositions = wellPositions
    self._hyperparameters = hyperparameters
    self._auDessusPerAnimalIdList = None
    
    self.cap = zzVideoReading.VideoCapture(self._videoPath, self._hyperparameters)
    if (self.cap.isOpened()== False):
      print("Error opening video stream or file")
    if ("readEventBasedData" in self._hyperparameters) and self._hyperparameters["readEventBasedData"]:
      self._hyperparameters["lastFrame"] = int(self.cap.get(cv2.CAP_PROP_FRAME_COUNT))
    
    self._firstFrame = self._hyperparameters["firstFrame"]
    self._lastFrame = self._hyperparameters["lastFrame"]
    self._nbTailPoints = self._hyperparameters["nbTailPoints"]
    self._previousFrames = None
    self._trackingDataPerWell = [np.zeros((self._hyperparameters["nbAnimalsPerWell"], self._lastFrame-self._firstFrame+1, self._nbTailPoints, 2)) for _ in range(len(self._wellPositions))]
    self._lastFirstTheta = np.zeros(len(self._wellPositions))
    self._lastFirstTheta[:] = -99999
    self._listOfWellsOnWhichToRunTheTracking = [i for i in range(0, len(self._wellPositions))] if hyperparameters["onlyTrackThisOneWell"] == -1 else [hyperparameters["onlyTrackThisOneWell"]]
    self._times2 = np.zeros((self._lastFrame - self._firstFrame + 1, 5))
    self._printInterTime = False

  def _debugFrame(self, frame, title=None, buttons=(), timeout=None):
    util.showFrame(frame, title=title, buttons=buttons, timeout=timeout)

  def _adjustParameters(self, i, frame, unprocessedFrame, widgets):
    return None

  def run(self):
    
    ### Step 1 (out of 2): Tracking:
    resizeFrameFactor = self._hyperparameters["resizeFrameFactor"] if "resizeFrameFactor" in self._hyperparameters else 0
    
    # Getting video reader
    cap = self.cap
    if (cap.isOpened()== False):
      print("Error opening video stream or file")
    
    # Simple background extraction with first and last frame of the video + Getting list of wells on which to run the tracking
    ret, self._background = cap.read()
    if resizeFrameFactor:
      self._background = cv2.resize(self._background, (int(len(self._background[0])/resizeFrameFactor), int(len(self._background)/resizeFrameFactor)))
    if "lastFrameForInitialBackDetect" in self._hyperparameters and self._hyperparameters["lastFrameForInitialBackDetect"]:
      if int(cap.get(cv2.CAP_PROP_FRAME_COUNT)) <= self._hyperparameters["lastFrameForInitialBackDetect"]:
        cap.set(cv2.CAP_PROP_POS_FRAMES, int(cap.get(cv2.CAP_PROP_FRAME_COUNT)) - 1)
      else:
        cap.set(cv2.CAP_PROP_POS_FRAMES, self._hyperparameters["lastFrameForInitialBackDetect"])
    else:
      cap.set(cv2.CAP_PROP_POS_FRAMES, int(cap.get(cv2.CAP_PROP_FRAME_COUNT)) - 1)
    ret, frame = cap.read()
    if resizeFrameFactor:
      frame = cv2.resize(frame, (int(len(frame[0])/resizeFrameFactor), int(len(frame)/resizeFrameFactor)))
    if self._hyperparameters["chooseWellsToRunTrackingOnWithFirstAndLastFrame"] and self._hyperparameters["onlyTrackThisOneWell"] == -1:
      self._listOfWellsOnWhichToRunTheTracking = getListOfWellsOnWhichToRunTheTracking(self, self._background[:,:,0], frame[:,:,0])
    print("listOfWellsOnWhichToRunTheTracking:", self._listOfWellsOnWhichToRunTheTracking)
    if not("noBackgroundSubtraction" in self._hyperparameters) or not(self._hyperparameters["noBackgroundSubtraction"]):
      if not(self._hyperparameters["useFirstFrameAsBackground"]):
        self._background = cv2.max(frame, self._background) # INCONSISTENT!!! should be changed!
      if len(self._background.shape) == 3 and self._background.shape[2] == 3:
        self._background = cv2.cvtColor(self._background, cv2.COLOR_BGR2GRAY) # INCONSISTENT!!! should be changed!
    cap.set(cv2.CAP_PROP_POS_FRAMES, self._firstFrame)
    
    if resizeFrameFactor:
      initialWellPositions = self._wellPositions.copy()
      self._wellPositions = [{'topLeftX': int(pos['topLeftX']/resizeFrameFactor), 'topLeftY': int(pos['topLeftY']/resizeFrameFactor), 'lengthX': int(pos['lengthX']/resizeFrameFactor), 'lengthY': int(pos['lengthY']/resizeFrameFactor)} for pos in self._wellPositions]
    
    # Initializing variables
    times  = np.zeros((self._lastFrame - self._firstFrame + 1, 2))
    ret = True
    
    # Going through each frame of the video
    widgets = None
    startTime = time.time()
    k = self._firstFrame
    while (ret and k <= self._lastFrame):
      if self._hyperparameters["freqAlgoPosFollow"] and k % self._hyperparameters["freqAlgoPosFollow"] == 0:
        print("Tracking at frame", k)
      time1 = time.time()
      if self._hyperparameters['adjustFreelySwimTracking']:
        cap.set(1, k)
      ret, frame = cap.read()
      time2 = time.time()
      if resizeFrameFactor:
        frame = cv2.resize(frame, (int(len(frame[0])/resizeFrameFactor), int(len(frame)/resizeFrameFactor)))
      if ret:
        if self._hyperparameters["backgroundSubtractionOnWholeImage"] or k == self._firstFrame:
          frameROI, unprocessedFrameROI = backgroundSubtractionOnWholeImage(self, frame, k-self._firstFrame)
        else:
          backgroundSubtractionOnlyOnROIs(self, frame, k-self._firstFrame)
        if self._hyperparameters["updateBackgroundAtInterval"]:
          for wellNumber in range(0, len(self._wellPositions)):
            self._updateBackgroundAtInterval(k, wellNumber, frame[self._wellPositions[wellNumber]["topLeftY"]:self._wellPositions[wellNumber]["topLeftY"]+self._wellPositions[wellNumber]["lengthY"], self._wellPositions[wellNumber]["topLeftX"]:self._wellPositions[wellNumber]["topLeftX"]+self._wellPositions[wellNumber]["lengthX"], 0], self._trackingDataPerWell[wellNumber], frame)
      
      time3 = time.time()
      times[k-self._firstFrame, 0] = time2 - time1
      times[k-self._firstFrame, 1] = time3 - time2
      adjustParamsInfo = self._adjustParameters(k, frameROI, unprocessedFrameROI, widgets)
      if adjustParamsInfo is not None:
        k, widgets = adjustParamsInfo
        if self._nbTailPoints != self._hyperparameters["nbTailPoints"]:
          self._nbTailPoints = self._hyperparameters["nbTailPoints"]
          self._trackingDataPerWell = [np.zeros((self._hyperparameters["nbAnimalsPerWell"], self._lastFrame-self._firstFrame+1, self._nbTailPoints, 2)) for _ in range(len(self._wellPositions))]
      else:
        k += 1
    
    
    if resizeFrameFactor:
      self._trackingDataPerWell = [resizeFrameFactor * elem for elem in self._trackingDataPerWell]
      self._wellPositions = initialWellPositions
    
    endTime = time.time()
    
    cap.release()
    
    if self._hyperparameters["detectMovementWithRawVideoInsideTracking"]:
      frameGapComparision = self._hyperparameters["frameGapComparision"]
      nbFrames = len(self._trackingDataPerWell[0][0])
      for wellNumber in range(len(self._trackingDataPerWell)):
        for animalId in range(len(self._trackingDataPerWell[wellNumber])):
          self._trackingDataPerWell[wellNumber][animalId][:nbFrames-frameGapComparision][:][:] = self._trackingDataPerWell[wellNumber][animalId][frameGapComparision:nbFrames][:][:]
          self._auDessusPerAnimalIdList[wellNumber][animalId][:nbFrames-frameGapComparision] = self._auDessusPerAnimalIdList[wellNumber][animalId][frameGapComparision:nbFrames]
    
    print("")
    print("Color to grey:"           , np.median(self._times2[:,0]))
    print("Bout detection:"          , np.median(self._times2[:,1]))
    print("Background substraction:" , np.median(self._times2[:,2]))
    print("Gaussian blur:"           , np.median(self._times2[:,3]))
    print("Tracking on each well:"   , np.median(self._times2[:,4]))
    
    loadingImagesTime       = np.median(times[:,0])
    processingImagesTime    = np.median(times[:,1])
    percentTimeSpentLoading = loadingImagesTime / (loadingImagesTime + processingImagesTime)
    print("Median time spent on: Loading images:", loadingImagesTime, "; Processing images:", processingImagesTime)
    print("Percentage of time spent loading images:", percentTimeSpentLoading*100)
    print("Total tracking Time:", endTime - startTime)
    print("Tracking Time (without loading image):", (endTime - startTime) * (1 - percentTimeSpentLoading))
    print("Total tracking fps:", k / (endTime - startTime))
    print("Tracking fps (without loading image):", k / ((endTime - startTime) * (1 - percentTimeSpentLoading)))
    print("")
    
    ### Step 2 (out of 2): Extracting bout of movements:
    
    if self._hyperparameters["detectMovementWithRawVideoInsideTracking"] and self._hyperparameters["thresForDetectMovementWithRawVideo"]:
    
      trackingHeadingAllAnimalsList = [[[((calculateAngle(self._trackingDataPerWell[wellNumber][animalNumber][i][0][0], self._trackingDataPerWell[wellNumber][animalNumber][i][0][1], self._trackingDataPerWell[wellNumber][animalNumber][i][1][0], self._trackingDataPerWell[wellNumber][animalNumber][i][1][1]) + math.pi) % (2 * math.pi) if len(self._trackingDataPerWell[wellNumber][0][i]) > 1 else 0) for i in range(0, self._lastFrame-self._firstFrame+1)] for animalNumber in range(0, self._hyperparameters["nbAnimalsPerWell"])] for wellNumber in range(0, len(self._wellPositions))]
      
      return {wellNumber: extractParameters([self._trackingDataPerWell[wellNumber], trackingHeadingAllAnimalsList[wellNumber], [], 0, 0, self._auDessusPerAnimalIdList[wellNumber]], wellNumber, self._hyperparameters, self._videoPath, self._wellPositions, self._background) for wellNumber in range(0, len(self._wellPositions))}

    elif self._hyperparameters["adjustDetectMovWithRawVideo"]:

      trackingHeadingAllAnimalsList = [[[((calculateAngle(self._trackingDataPerWell[wellNumber][animalNumber][i][0][0], self._trackingDataPerWell[wellNumber][animalNumber][i][0][1], self._trackingDataPerWell[wellNumber][animalNumber][i][1][0], self._trackingDataPerWell[wellNumber][animalNumber][i][1][1]) + math.pi) % (2 * math.pi) if len(self._trackingDataPerWell[wellNumber][0][i]) > 1 else 0) for i in range(0, self._lastFrame-self._firstFrame+1)] for animalNumber in range(0, self._hyperparameters["nbAnimalsPerWell"])] for wellNumber in range(0, len(self._wellPositions))]

      return {wellNumber: extractParameters([self._trackingDataPerWell[wellNumber], trackingHeadingAllAnimalsList[wellNumber], [], 0, 0], wellNumber, self._hyperparameters, self._videoPath, self._wellPositions, self._background) for wellNumber in self._listOfWellsOnWhichToRunTheTracking}

    elif  self._hyperparameters["coordinatesOnlyBoutDetection"]:

      trackingHeadingAllAnimalsList = [[[((calculateAngle(self._trackingDataPerWell[wellNumber][animalNumber][i][0][0], self._trackingDataPerWell[wellNumber][animalNumber][i][0][1], self._trackingDataPerWell[wellNumber][animalNumber][i][1][0], self._trackingDataPerWell[wellNumber][animalNumber][i][1][1]) + math.pi) % (2 * math.pi) if len(self._trackingDataPerWell[wellNumber][0][i]) > 1 else 0) for i in range(0, self._lastFrame-self._firstFrame+1)] for animalNumber in range(0, self._hyperparameters["nbAnimalsPerWell"])] for wellNumber in range(0, len(self._wellPositions))]

      return {wellNumber: extractParameters([self._trackingDataPerWell[wellNumber], trackingHeadingAllAnimalsList[wellNumber], [], 0, 0], wellNumber, self._hyperparameters, self._videoPath, self._wellPositions, self._background) for wellNumber in range(0, len(self._wellPositions))}

    else:
    
      if not(self._hyperparameters["detectBouts"]):
        
        trackingHeadingAllAnimalsList = [[[((calculateAngle(self._trackingDataPerWell[wellNumber][animalNumber][i][0][0], self._trackingDataPerWell[wellNumber][animalNumber][i][0][1], self._trackingDataPerWell[wellNumber][animalNumber][i][1][0], self._trackingDataPerWell[wellNumber][animalNumber][i][1][1]) + math.pi) % (2 * math.pi) if len(self._trackingDataPerWell[wellNumber][0][i]) > 1 else 0) for i in range(0, self._lastFrame-self._firstFrame+1)] for animalNumber in range(0, self._hyperparameters["nbAnimalsPerWell"])] for wellNumber in range(0, len(self._wellPositions))]
        
        self._auDessusPerAnimalIdList = [[np.ones((self._lastFrame-self._firstFrame+1, 1)) for nbAnimalsPerWell in range(0, self._hyperparameters["nbAnimalsPerWell"])] for wellNumber in range(len(self._wellPositions))]
        
        return {wellNumber: extractParameters([self._trackingDataPerWell[wellNumber], trackingHeadingAllAnimalsList[wellNumber], [], 0, 0, self._auDessusPerAnimalIdList[wellNumber]], wellNumber, self._hyperparameters, self._videoPath, self._wellPositions, self._background) for wellNumber in range(0, len(self._wellPositions))}
        
      else:
        
        outputData = detectMovementWithTrackedDataAfterTracking(self)
      
      return outputData


zebrazoom.code.tracking.register_tracking_method('fastFishTracking.tracking', Tracking)



import math
import cv2

class UpdateBackgroundAtIntervalMixin:
  def _updateBackgroundAtInterval(self, i, wellNumber, initialCurFrame, trackingHeadTailAllAnimals, frame):
    if i % self._hyperparameters["updateBackgroundAtInterval"] == 0:
      showImages = False
      firstFrameToShow = -1
      if showImages and i > firstFrameToShow:
        self._debugFrame(self._background, title='background before')
      xvalues = [trackingHeadTailAllAnimals[0, i-self._firstFrame][k][0] for k in range(0, len(trackingHeadTailAllAnimals[0, i-self._firstFrame]))]
      yvalues = [trackingHeadTailAllAnimals[0, i-self._firstFrame][k][1] for k in range(0, len(trackingHeadTailAllAnimals[0, i-self._firstFrame]))]
      xmin = min(xvalues)
      xmax = max(xvalues)
      ymin = min(yvalues)
      ymax = max(yvalues)
      dist = 1 * math.sqrt((xmax - xmin) ** 2 + (ymax - ymin) ** 2)
      xmin = int(xmin - dist) if xmin - dist >= 0 else 0
      xmax = int(xmax + dist) if xmax + dist < len(frame[0]) else len(frame[0]) - 1
      ymin = int(ymin - dist) if ymin - dist >= 0 else 0
      ymax = int(ymax + dist) if ymax + dist < len(frame) else len(frame) - 1
      if xmin != xmax and ymin != ymax:
        partOfBackgroundToSave = self._background[self._wellPositions[wellNumber]["topLeftY"]+ymin:self._wellPositions[wellNumber]["topLeftY"]+ymax, self._wellPositions[wellNumber]["topLeftX"]+xmin:self._wellPositions[wellNumber]["topLeftX"]+xmax].copy() # copy ???
        if showImages and i > firstFrameToShow:
          self._debugFrame(partOfBackgroundToSave, title='partOfBackgroundToSave')
      self._background[self._wellPositions[wellNumber]["topLeftY"]:self._wellPositions[wellNumber]["topLeftY"]+self._wellPositions[wellNumber]["lengthY"], self._wellPositions[wellNumber]["topLeftX"]:self._wellPositions[wellNumber]["topLeftX"]+self._wellPositions[wellNumber]["lengthX"]] = initialCurFrame.copy()
      if showImages and i > firstFrameToShow:
        self._debugFrame(self._background, title='background middle')
      if xmin != xmax and ymin != ymax:
        self._background[self._wellPositions[wellNumber]["topLeftY"]+ymin:self._wellPositions[wellNumber]["topLeftY"]+ymax, self._wellPositions[wellNumber]["topLeftX"]+xmin:self._wellPositions[wellNumber]["topLeftX"]+xmax] = partOfBackgroundToSave
      if showImages and i > firstFrameToShow:
        self._debugFrame(self._background, title='background after')

  def _updateBackgroundAtIntervalRememberLastFrame(self, i, wellNumber, initialCurFrame, lastFrameRememberedForBackgroundExtract):
    if (i - self._firstFrame) % self._hyperparameters["updateBackgroundAtIntervalRememberLastFrame"] == 0:
      if i != self._firstFrame:
        if self._hyperparameters["extractBackWhiteBackground"]:
          self._background = cv2.max(initialCurFrame, lastFrameRememberedForBackgroundExtract)
        else:
          self._background = cv2.min(initialCurFrame, lastFrameRememberedForBackgroundExtract)
      return initialCurFrame
    else:
      return lastFrameRememberedForBackgroundExtract



from pathlib import Path
import os
import json
import numpy as np
import math
import cv2


def drawWhitePointsOnInitialImages(initialCurFrame, back, hyperparameters):
  import zebrazoom.code.util as util

  img = util.drawPoints(initialCurFrame.copy(), "Draw white points and click on any key when you're done")
  putToWhite = ( img.astype('int32') >= (back.astype('int32') - hyperparameters["minPixelDiffForBackExtract"]) )
  img[putToWhite] = 255      
  ret, thresh1 = cv2.threshold(img, hyperparameters["thresholdForBlobImg"], 255, cv2.THRESH_BINARY)
  erodeSize = hyperparameters["erodeSize"]
  kernel    = np.ones((erodeSize,erodeSize), np.uint8)
  thresh1   = cv2.erode(thresh1, kernel, iterations=hyperparameters["erodeIter"])
  
  return [img, thresh1]


def saveImagesAndData(hyperparameters, bodyContour, initialCurFrame, wellNumber, frameNumber):
  
  if hyperparameters["saveBodyMaskResampleContourNbPoints"]:
    bodyContour2  = bodyContour.copy()
    contourLength = 0
    for contourInd in range(len(bodyContour) - 2):
      pt1 = bodyContour[contourInd][0]
      pt2 = bodyContour[contourInd+1][0]
      contourLength = contourLength + math.sqrt((pt1[0] - pt2[0])**2 + (pt1[1] - pt2[1])**2)
    step = contourLength / hyperparameters["saveBodyMaskResampleContourNbPoints"]
    curLength = 0
    curPoint  = 0
    indexToDelete = []
    for contourInd in range(len(bodyContour) - 2):
      pt1 = bodyContour[contourInd][0]
      pt2 = bodyContour[contourInd+1][0]
      curLength = curLength + math.sqrt((pt1[0] - pt2[0])**2 + (pt1[1] - pt2[1])**2)
      if curLength >= curPoint * step:
        curPoint = curPoint + 1
      else:
        indexToDelete.append(contourInd)
    indexToDelete.reverse()
    
    for indToDelete in indexToDelete:
      bodyContour2 = np.delete(bodyContour2, indToDelete, 0)
    bodyContour = bodyContour2
  
  if type(bodyContour) != int and len(bodyContour) and type(initialCurFrame) != int:
    initialCurFrame2 = initialCurFrame.copy()
    initialCurFrame2 = cv2.cvtColor(initialCurFrame2, cv2.COLOR_GRAY2RGB)
    lenX = len(initialCurFrame2[0])
    lenY = len(initialCurFrame2)
    for contourPt in bodyContour:
      pt1 = contourPt[0]
      initialCurFrame2 = cv2.circle(initialCurFrame2, (pt1[0], pt1[1]), 1, (255, 0, 0), -1)
    
    if hyperparameters["bodyMask_saveDataForAllFrames"]:
      answerYes = True        
    else:
      import zebrazoom.code.util as util

      answerYes = False

      def saveClicked():
        nonlocal answerYes
        answerYes = True
      buttons = (("Save", saveClicked), ("Discard", None))
      util.showFrame(initialCurFrame2, title="Is this a good delimitation of the body of the animal?", buttons=buttons)
    
    if answerYes:
      
      if hyperparameters["bodyMask_saveAsPngMask"]:
        originalShape = np.zeros((lenY, lenX))
        originalShape[:, :] = 0
        originalShape = originalShape.astype(np.uint8)
        cv2.fillPoly(originalShape, pts =[bodyContour], color=(1))
      
      pathToImg = os.path.join(hyperparameters["outputFolder"], hyperparameters["videoNameWithTimestamp"])
      imgName   = hyperparameters["videoName"] + '_well' + str(wellNumber) + '_frame' + str(frameNumber)
      
      if not os.path.exists(os.path.join(pathToImg, 'PNGMasks')):
        os.makedirs(os.path.join(pathToImg, 'PNGMasks'))
      if not os.path.exists(os.path.join(pathToImg, 'PNGImages')):
        os.makedirs(os.path.join(pathToImg, 'PNGImages'))
      
      cv2.imwrite(os.path.join(os.path.join(pathToImg, 'PNGImages'), imgName + '.png'), initialCurFrame)
      if hyperparameters["bodyMask_saveAsPngMask"]:
        cv2.imwrite(os.path.join(os.path.join(pathToImg, 'PNGMasks'),  imgName + '_mask.png'), originalShape)
      
      if hyperparameters["bodyMask_saveAsLabelMeJsonFormat"]:
        # Labels can then be reloaded and changed with the command: labelme --nodata
        
        jsonData = '''
            {
              "version": "4.5.9",
              "flags": {},
              "shapes": [
                {
                  "label": "Person",
                  "group_id": null,
                  "shape_type": "polygon",
                  "flags": {}
                }],
              "imageData": null
            }
            '''
        jsonData = json.loads(jsonData)
        jsonData['shapes'][0]['points'] = [elem[0].tolist() for elem in bodyContour]
        jsonData['imagePath']   = imgName + ".png"
        jsonData['imageHeight'] = lenY
        jsonData['imageWidth']  = lenX
        
        with open(os.path.join(os.path.join(pathToImg, 'PNGImages'), imgName + '.json'), 'w') as outfile:
          json.dump(jsonData, outfile)


def createMask(pathToImgFolder):
  
  pathToMaskDir = Path(pathToImgFolder)
  pathToMaskDir = pathToMaskDir.parent
  pathToMaskDir = os.path.join(pathToMaskDir, 'PNGMasks')
  if not(os.path.exists(pathToMaskDir)):
    os.mkdir(pathToMaskDir)
  
  os.walk(pathToImgFolder)
  for x in sorted(next(os.walk(pathToImgFolder))[2]):
    if '.json' in x:
      with open(os.path.join(pathToImgFolder, x)) as json_file:
        data = json.load(json_file)
        for i in range(0, len(data["shapes"])):
          bodyContour   = np.array([data["shapes"][i]["points"]])
          originalShape = np.zeros((data["imageHeight"], data["imageWidth"]))
          originalShape[:, :] = 0
          originalShape = originalShape.astype(np.uint8)
          bodyContour = bodyContour.astype(np.int)
          cv2.fillPoly(originalShape, pts =[bodyContour], color=(1))
          cv2.imwrite(os.path.join(pathToMaskDir, x[:len(x)-5] + '_mask.png'), originalShape)






